# automatically generated by the FlatBuffers compiler, do not modify

# namespace: Dl

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class MapType(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = MapType()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsMapType(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # MapType
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # MapType
    def KeyType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # MapType
    def ValueType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from FlatBuffers.Dl.TypeInfo import TypeInfo
            obj = TypeInfo()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def MapTypeStart(builder):
    builder.StartObject(2)

def Start(builder):
    MapTypeStart(builder)

def MapTypeAddKeyType(builder, keyType):
    builder.PrependInt32Slot(0, keyType, 0)

def AddKeyType(builder, keyType):
    MapTypeAddKeyType(builder, keyType)

def MapTypeAddValueType(builder, valueType):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(valueType), 0)

def AddValueType(builder, valueType):
    MapTypeAddValueType(builder, valueType)

def MapTypeEnd(builder):
    return builder.EndObject()

def End(builder):
    return MapTypeEnd(builder)

import FlatBuffers.Dl.TypeInfo
try:
    from typing import Optional
except:
    pass

class MapTypeT(object):

    # MapTypeT
    def __init__(self):
        self.keyType = 0  # type: int
        self.valueType = None  # type: Optional[FlatBuffers.Dl.TypeInfo.TypeInfoT]

    @classmethod
    def InitFromBuf(cls, buf, pos):
        mapType = MapType()
        mapType.Init(buf, pos)
        return cls.InitFromObj(mapType)

    @classmethod
    def InitFromPackedBuf(cls, buf, pos=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, pos)
        return cls.InitFromBuf(buf, pos+n)

    @classmethod
    def InitFromObj(cls, mapType):
        x = MapTypeT()
        x._UnPack(mapType)
        return x

    # MapTypeT
    def _UnPack(self, mapType):
        if mapType is None:
            return
        self.keyType = mapType.KeyType()
        if mapType.ValueType() is not None:
            self.valueType = FlatBuffers.Dl.TypeInfo.TypeInfoT.InitFromObj(mapType.ValueType())

    # MapTypeT
    def Pack(self, builder):
        if self.valueType is not None:
            valueType = self.valueType.Pack(builder)
        MapTypeStart(builder)
        MapTypeAddKeyType(builder, self.keyType)
        if self.valueType is not None:
            MapTypeAddValueType(builder, valueType)
        mapType = MapTypeEnd(builder)
        return mapType



from typing import Iterable

import torch
import torchvision

from esp_ppq import TargetPlatform, graphwise_error_analyse
from esp_ppq.api import QuantizationSettingFactory, quantize_torch_model
from esp_ppq.api.interface import ENABLE_CUDA_KERNEL
from esp_ppq.executor.torch import TorchExecutor

# ------------------------------------------------------------
# 在 PPQ 中我们目前提供两种不同的算法帮助你微调网络
# 这些算法将使用 calibration dataset 中的数据，对网络权重展开重训练
# 1. 经过训练的网络不保证中间结果与原来能够对齐，在进行误差分析时你需要注意这一点
# 2. 在训练中使用 with ENABLE_CUDA_KERNEL(): 子句将显著加速训练过程
# 3. 训练过程的缓存数据将被贮存在 gpu 上，这可能导致你显存溢出，你可以修改参数将缓存设备改为 cpu
# ------------------------------------------------------------

BATCHSIZE = 32
INPUT_SHAPE = [BATCHSIZE, 3, 224, 224]
DEVICE = 'cuda'
PLATFORM = TargetPlatform.PPL_CUDA_INT8


def load_calibration_dataset() -> Iterable:
    # ------------------------------------------------------------
    # 让我们从创建 calibration 数据开始做起， PPQ 需要你送入 32 ~ 1024 个样本数据作为校准数据集
    # 它们应该尽可能服从真实样本的分布，量化过程如同训练过程一样存在可能的过拟合问题
    # 你应当保证校准数据是经过正确预处理的、有代表性的数据，否则量化将会失败；校准数据不需要标签；数据集不能乱序
    # ------------------------------------------------------------
    return [torch.rand(size=INPUT_SHAPE) for _ in range(32)]


CALIBRATION = load_calibration_dataset()


def collate_fn(batch: torch.Tensor) -> torch.Tensor:
    return batch.to(DEVICE)


# ------------------------------------------------------------
# 我们使用 mobilenet v2 作为一个样例模型
# PPQ 将会使用 torch.onnx.export 函数 把 pytorch 的模型转换为 onnx 模型
# 对于复杂的 pytorch 模型而言，你或许需要自己完成 pytorch 模型到 onnx 的转换过程
# ------------------------------------------------------------
model = torchvision.models.mobilenet.mobilenet_v2(pretrained=True)
model = model.to(DEVICE)

# ------------------------------------------------------------
# PPQ 提供基于 LSQ 的网络微调过程，这是推荐的做法
# 你将使用 Quant Setting 来调用微调过程，并调整微调参数
# ------------------------------------------------------------
QSetting = QuantizationSettingFactory.default_setting()
QSetting.lsq_optimization = True
QSetting.lsq_optimization_setting.block_size = 4
QSetting.lsq_optimization_setting.lr = 1e-5
QSetting.lsq_optimization_setting.gamma = 0
QSetting.lsq_optimization_setting.is_scale_trainable = True
QSetting.lsq_optimization_setting.collecting_device = 'cuda'

# ------------------------------------------------------------
# 如果你使用 ENABLE_CUDA_KERNEL 方法
# PPQ 将会尝试编译自定义的高性能量化算子，这一过程需要编译环境的支持
# 如果你在编译过程中发生错误，你可以删除此处对于 ENABLE_CUDA_KERNEL 方法的调用
# 这将显著降低 PPQ 的运算速度；但即使你无法编译这些算子，你仍然可以使用 pytorch 的 gpu 算子完成量化
# ------------------------------------------------------------
with ENABLE_CUDA_KERNEL():
    quantized = quantize_torch_model(
        model=model,
        calib_dataloader=CALIBRATION,
        calib_steps=32,
        input_shape=INPUT_SHAPE,
        setting=QSetting,
        collate_fn=collate_fn,
        platform=PLATFORM,
        onnx_export_file='Output/model.onnx',
        device=DEVICE,
        verbose=0,
    )

    # ------------------------------------------------------------
    # 当我们完成训练后，我们将调用 graphwise_error_analyse 方法分析网络误差
    # 经过训练的中间层误差可能很大，但这不是我们所关心的 —— 训练方法只优化最终输出的误差
    # 一个量化良好的网络，最后输出层的误差不应大于 10%
    # ------------------------------------------------------------
    graphwise_error_analyse(graph=quantized, running_device=DEVICE, dataloader=CALIBRATION, collate_fn=collate_fn)


model = torchvision.models.mobilenet.mobilenet_v2(pretrained=True)
model = model.to(DEVICE)

QSetting = QuantizationSettingFactory.default_setting()

with ENABLE_CUDA_KERNEL():
    quantized = quantize_torch_model(
        model=model,
        calib_dataloader=CALIBRATION,
        calib_steps=32,
        input_shape=INPUT_SHAPE,
        setting=QSetting,
        collate_fn=collate_fn,
        platform=PLATFORM,
        onnx_export_file='Output/model.onnx',
        device=DEVICE,
        verbose=0,
    )

    graphwise_error_analyse(graph=quantized, running_device=DEVICE, dataloader=CALIBRATION, collate_fn=collate_fn)



from .AscendQuantizer import AscendQuantizer
from .base import BaseQuantizer
from .DSPQuantizer import PPL_DSP_Quantizer, PPL_DSP_TI_Quantizer
from .EspdlQuantizer import (
    EspdlCHPreInt16Quantizer,
    EspdlCInt16Quantizer,
    EspdlCQuantizer,
    EspdlHPreInt16Quantizer,
    EspdlInt16Quantizer,
    EspdlQuantizer,
    EspdlS3HPreInt16Quantizer,
    EspdlS3Int16Quantizer,
    EspdlS3Quantizer,
)
from .FP8Quantizer import GraphCoreQuantizer, TensorRTQuantizer_FP8

# from .TRTQuantizer import TensorRTQuantizer
from .FPGAQuantizer import FPGAQuantizer
from .MetaxQuantizer import MetaxChannelwiseQuantizer, MetaxTensorwiseQuantizer
from .MNNQuantizer import MNNQuantizer
from .MyQuantizer import ExtQuantizer
from .NCNNQuantizer import NCNNQuantizer
from .NXPQuantizer import NXP_Quantizer
from .OpenvinoQuantizer import OpenvinoQuantizer
from .ORTQuantizer import OnnxruntimeQuantizer
from .PPLQuantizer import PPLCUDAQuantizer
from .RKNNQuantizer import RKNN_PerChannelQuantizer, RKNN_PerTensorQuantizer
from .TengineQuantizer import TengineQuantizer
from .TensorRTQuantizer import TensorRTQuantizer, TensorRTQuantizer_InputOnly



import os
import shutil
from typing import Iterable, List, Union

import numpy as np
import torch
from tqdm import tqdm

from esp_ppq.core import *
from esp_ppq.core import ppq_warning
from esp_ppq.executor import TorchExecutor
from esp_ppq.IR import BaseGraph, QuantableVariable
from esp_ppq.quantization.analyse.util import MeasurePrinter, MeasureRecorder
from esp_ppq.utils.fetch import tensor_random_fetch


def load_calibration_dataset(
    directory: str, input_shape: List[int], batchsize: int, input_format: str = 'chw'
) -> Iterable:
    """使用这个函数来加载校准数据集，校准数据集将被用来量化你的模型。这个函数只被用来加载图像数据集。
    你需要给出校准数据集位置，我们建议你将每张图片都单独保存到文件中，这个函数会自己完成后续的打包处理工作。
    校准数据集不应过大，这个函数会将所有数据加载到内存中，同时过大的校准数据集也不利于后续的量化处理操作。

    我们推荐你使用 512 ~ 4096 张图片进行校准，batchsize 设置为 16 ~ 64。
    我们支持读入 .npy 格式的数据，以及 .bin 或 .raw 的二进制数据，如果你选择以二进制格式输入数据，则必须指定样本尺寸
    如果你的样本尺寸不一致（即动态尺寸输入），则你必须使用 .npy 格式保存你的数据。

    如果这个函数无法满足你的需求，例如你的模型存在多个输入，则你可以自行构建数据集
    ppq 支持使用任何可遍历对象充当数据集，包括 torch.Dataset, list 等。

    Args:
        directory (str): 加载数据集的目录，目录不应包含子文件夹，所有目录中的文件将被视为数据。
        input_shape (List[int]): 图像尺寸，对于二进制输入文件而言，你必须指定图像尺寸，对于 npy文件 此项不起作用
        batchsize (int): batchsize 大小，这个函数会自动进行打包操作，但是如果你的数据本身已经有了预设的batchsize，
            则该函数不会覆盖原有batchsize
        input_format (str, optional): chw 或 hwc，指定输入图像数据排布。即使你的图像具有batch维度，它仍然将正常工作。

    Raises:
        FileNotFoundError: _description_
        ValueError: _description_

    Returns:
        Iterable: _description_
    """

    if not os.path.exists(directory) or not os.path.isdir(directory):
        raise FileNotFoundError(f'无法从指定位置加载数据集 {directory}. 目录不存在或无法访问，检查你的输入路径')
    if input_format not in {'chw', 'hwc'}:
        raise ValueError(f'无法理解的数据格式，对于图片数据，数据格式只能是 chw 或 hwc，而你输入了 {input_format}')

    num_of_file, samples, sizes = 0, [], set()
    for file in os.listdir(os.path.join(directory, 'data')):
        sample = None
        if file.endswith('.npy'):
            sample = np.load(os.path.join(directory, 'data', file))
            num_of_file += 1
        elif file.endswith('.bin') or file.endswith('.raw'):
            sample = np.fromfile(os.path.join(directory, 'data', file), dtype=np.float32)
            assert isinstance(sample, np.ndarray), f'数据应当是 numpy.ndarray，然而你输入了 {type(sample)}'
            sample = sample.reshape(input_shape)
            num_of_file += 1
        else:
            ppq_warning(f'文件格式不可读: {os.path.join(directory, "data", file)}, 该文件已经被忽略.')

        sample = convert_any_to_torch_tensor(sample)
        sample = sample.float()

        if sample.ndim == 3:
            sample = sample.unsqueeze(0)
        if input_format == 'hwc':
            sample = sample.permute([0, 3, 1, 2])

        if sample.shape[1] not in {1, 3, 4}:
            ppq_warning(f'Quantized data is not a regular channel, which channel is {sample.shape[1]}.')

        assert sample.shape[0] == 1 or batchsize == 1, (
            '你的输入图像似乎已经有了预设好的 batchsize, 因此我们不会再尝试对你的输入进行打包。'
        )

        sizes.add((sample.shape[-2], sample.shape[-1]))
        samples.append(sample)

    if len(sizes) != 1:
        ppq_warning('你的输入图像似乎包含动态的尺寸，因此 CALIBRATION BATCHSIZE 被强制设置为 1')
        batchsize = 1

    # create batches
    batches, batch = [], []
    if batchsize != 1:
        for sample in samples:
            if len(batch) < batchsize:
                batch.append(sample)
            else:
                batches.append(torch.cat(batch, dim=0))
                batch = [sample]
        if len(batch) != 0:
            batches.append(torch.cat(batch, dim=0))
    else:
        batches = samples

    print(f'{num_of_file} File(s) Loaded.')
    for idx, tensor in enumerate(samples[:5]):
        print(f'Loaded sample {idx}, shape: {tensor.shape}')
    assert len(batches) > 0, '你送入了空的数据集'

    print(f'Batch Shape: {batches[0].shape}')
    return batches


def load_from_file(
    filename: str, dtype: np.dtype = np.float32, shape: List[int] = None, format: str = '.npy'
) -> np.ndarray:
    if not os.path.exists(filename) or os.path.isdir(filename):
        raise FileNotFoundError(f'Can not load data from file: {filename}. File not exist or it is a directory.')

    if format == '.npy':
        return np.load(filename)
    elif format == '.dat':
        raw = np.fromfile(file=filename, dtype=dtype)
        if shape is not None:
            raw = np.reshape(raw, shape)
        return raw
    else:
        raise ValueError(f'file format not understand, support .npy and .dat only, {format} was given.')


def dump_to_file(filename: str, data: Union[torch.Tensor, np.ndarray], format: str = '.npy') -> None:
    if os.path.isdir(filename):
        raise FileExistsError(f'Can not dump data to file: {filename}', ' Cause it is a directory.')
    if os.path.exists(filename):
        ppq_warning(f'Overwriting file {filename} ...')

    data = convert_any_to_numpy(data)
    if format == '.npy':
        np.save(file=filename, arr=data)
    elif format == '.dat':
        data.tofile(os.path.join(filename))
    else:
        raise ValueError(f'file format not understand, support .npy and .dat only, {format} was given.')


def create_dir(dir: str):
    try:
        os.mkdir(dir)
    except Exception as e:
        pass
    finally:
        if not os.path.isdir(dir):
            raise FileNotFoundError(f'Can not create directory at {dir}')


def compare_cosine_similarity_between_results(ref_dir: str, target_dir: str, method: str = 'cosine'):
    print(f'正从 {ref_dir} 与 {target_dir} 加载数据 ...')
    if not os.path.exists(ref_dir) or not os.path.isdir(ref_dir):
        raise FileNotFoundError(f'找不到路径 {ref_dir}')
    if not os.path.exists(target_dir) or not os.path.isdir(target_dir):
        raise FileNotFoundError(f'找不到路径 {target_dir}')

    samples, outputs = [], []
    for sample_name in os.listdir(target_dir):
        if os.path.isdir(os.path.join(target_dir, sample_name)):
            samples.append(sample_name)

    for output in os.listdir(os.path.join(target_dir, samples[-1])):
        outputs.append(output)

    print(f'总计 {len(samples)} 个 Sample 被检出，{len(outputs)} 个待测输出 Variable')
    recorders = {name: MeasureRecorder(measurement=method) for name in outputs}

    for sample_name in samples:
        for output in outputs:
            ref_file = os.path.join(ref_dir, sample_name, output)
            target_file = os.path.join(target_dir, sample_name, output)

            ref = load_from_file(filename=ref_file, format='.dat')[:1000]
            target = load_from_file(filename=target_file, format='.dat')[:1000]

            recorders[output].update(
                y_real=convert_any_to_torch_tensor(ref), y_pred=convert_any_to_torch_tensor(target)
            )

    results = {}
    for output, recorder in recorders.items():
        results[output] = recorder.measure

    method_str = 'MEASUREMENT'
    if method == 'snr':
        method_str = 'NOISE:SIGNAL POWER RATIO'
    if method == 'cosine':
        method_str = 'COSINE SIMILARITY'
    if method == 'mse':
        method_str = 'MSE LOSS(UNSCALED)'
    MeasurePrinter(results, order='large_to_small', measure=method_str, percentage=method in {'snr', 'cosine'}).print()


def dump_internal_results(
    graph: BaseGraph, dataloader: Iterable, dump_dir: str, executing_device: str, sample: bool = True
):
    i_dir = os.path.join(dump_dir, 'inputs')
    o_dir = os.path.join(dump_dir, 'outputs')

    create_dir(i_dir)
    create_dir(o_dir)

    # 找出所有量化点，抽出所有中间结果.
    for var in graph.variables.values():
        if isinstance(var, QuantableVariable):
            if var.source_op_config is not None and var.source_op_config.state == QuantizationStates.ACTIVATED:
                graph.outputs[var.name] = var  # 直接标记为网络输出

    executor = TorchExecutor(graph, device=executing_device)
    for batch_idx, batch in tqdm(enumerate(dataloader), total=len(dataloader), desc='Dumping Results ...'):
        batch = batch.to(executing_device)
        outputs = executor.forward(batch)

        create_dir(os.path.join(o_dir, str(batch_idx)))
        for name, output in zip(graph.outputs, outputs):
            # 如果数字太多就抽样
            if output.numel() > 10000 and sample:
                output = tensor_random_fetch(
                    tensor=output,
                    seed=10086,  # 保证随机种子一致才能比对结果
                    num_of_fetches=4096,
                )

            dump_to_file(filename=os.path.join(o_dir, str(batch_idx), name + '.dat'), data=output, format='.dat')

        dump_to_file(filename=os.path.join(i_dir, str(batch_idx) + '.npy'), data=batch, format='.npy')


def split_result_to_directory(raw_dir: str, to_dir: str):
    print(f'正从 {raw_dir} 分割数据')
    data_files, sample_names = [], set()
    for file_name in os.listdir(raw_dir):
        if file_name.endswith('.dat'):
            sample_name = file_name.split('_')[0]
            sample_names.add(sample_name)
            data_files.append((sample_name, file_name))

    print(f'加载 {len(data_files)} 数据文件，总计 {len(sample_names)} 个样本')
    create_dir(to_dir)
    for sample_name in sample_names:
        create_dir(os.path.join(to_dir, sample_name))
    for sample_name, file_name in tqdm(data_files, total=len(data_files), desc='正复制文件中 ...'):
        file_path = os.path.join(raw_dir, file_name)
        processed = os.path.join(to_dir, sample_name, file_name[file_name.index('_') + 1 :])
        shutil.copy(file_path, processed)



from typing import Iterable

import torch
import torchvision

from esp_ppq import TargetPlatform, graphwise_error_analyse
from esp_ppq.api.interface import ENABLE_CUDA_KERNEL, dump_torch_to_onnx, load_onnx_graph, quantize_native_model
from esp_ppq.api.setting import QuantizationSettingFactory

# ------------------------------------------------------------
# 在 PPQ 中我们提供许多校准方法，这些校准方法将计算出网络的量化参数
# 这个脚本将以随机数据和 mobilenet v2 网络为例向你展示它们的使用方法
# ------------------------------------------------------------

BATCHSIZE = 32
INPUT_SHAPE = [BATCHSIZE, 3, 224, 224]
DEVICE = 'cuda'
PLATFORM = TargetPlatform.TRT_INT8


def load_calibration_dataset() -> Iterable:
    # ------------------------------------------------------------
    # 让我们从创建 calibration 数据开始做起， PPQ 需要你送入 32 ~ 1024 个样本数据作为校准数据集
    # 它们应该尽可能服从真实样本的分布，量化过程如同训练过程一样存在可能的过拟合问题
    # 你应当保证校准数据是经过正确预处理的、有代表性的数据，否则量化将会失败；校准数据不需要标签；数据集不能乱序
    # ------------------------------------------------------------
    return [torch.rand(size=INPUT_SHAPE) for _ in range(32)]


CALIBRATION = load_calibration_dataset()


def collate_fn(batch: torch.Tensor) -> torch.Tensor:
    return batch.to(DEVICE)


# ------------------------------------------------------------
# 我们使用 mobilenet v2 作为一个样例模型
# PPQ 将会使用 torch.onnx.export 函数 把 pytorch 的模型转换为 onnx 模型
# 对于复杂的 pytorch 模型而言，你或许需要自己完成 pytorch 模型到 onnx 的转换过程
# ------------------------------------------------------------
model = torchvision.models.mobilenet.mobilenet_v2(pretrained=True)
model = model.to(DEVICE)

# ------------------------------------------------------------
# PPQ 提供 kl, mse, minmax, isotone, percentile(默认) 五种校准方法
# 每一种校准方法还有更多参数可供调整，PPQ 也允许你单独调整某一层的量化校准方法
# 在这里我们首先展示以 QSetting 的方法调整量化校准参数(推荐)
# ------------------------------------------------------------
QSetting = QuantizationSettingFactory.default_setting()
QSetting.quantize_activation_setting.calib_algorithm = 'kl'
QSetting.quantize_parameter_setting.calib_algorithm = 'minmax'
# ------------------------------------------------------------
# 更进一步地，当你选择了某种校准方法，你可以进入 esp_ppq.core.common
# OBSERVER_KL_HIST_BINS, OBSERVER_PERCENTILE, OBSERVER_MSE_HIST_BINS 皆是与校准方法相关的可调整参数
# OBSERVER_KL_HIST_BINS - KL 算法相关的箱子个数，你可以试试将其调整为 512, 1024, 2048, 4096, 8192 ...
# OBSERVER_PERCENTILE - Percentile 算法相关的百分比，你可以试试将其调整为 0.9999, 0.9995, 0.99999, 0.99995 ...
# OBSERVER_MSE_HIST_BINS - MSE 算法相关的箱子个数，你可以试试将其调整为 512, 1024, 2048, 4096, 8192 ...
# ------------------------------------------------------------

# ------------------------------------------------------------
# 如果你使用 ENABLE_CUDA_KERNEL 方法
# PPQ 将会尝试编译自定义的高性能量化算子，这一过程需要编译环境的支持
# 如果你在编译过程中发生错误，你可以删除此处对于 ENABLE_CUDA_KERNEL 方法的调用
# 这将显著降低 PPQ 的运算速度；但即使你无法编译这些算子，你仍然可以使用 pytorch 的 gpu 算子完成量化
# ------------------------------------------------------------
with ENABLE_CUDA_KERNEL():
    dump_torch_to_onnx(
        model=model, onnx_export_file='Output/model.onnx', input_shape=INPUT_SHAPE, input_dtype=torch.float32
    )
    graph = load_onnx_graph(onnx_import_file='Output/model.onnx')
    quantized = quantize_native_model(
        model=graph,
        calib_dataloader=CALIBRATION,
        calib_steps=32,
        input_shape=INPUT_SHAPE,
        collate_fn=collate_fn,
        platform=PLATFORM,
        device=DEVICE,
        verbose=0,
    )

    # ------------------------------------------------------------
    # graphwise_error_analyse 是最常用的分析方法，它分析网络中的量化误差情况，它的结果将直接打印在屏幕上
    # 对于 graphwise_error_analyse 而言，算子的误差直接衡量了量化网络与浮点网络之间的输出误差
    # 这一误差是累积的，意味着网络后面的算子总是会比网络前面的算子拥有更高的输出误差
    # 留意网络输出的误差情况，如果你想获得一个精度较高的量化网络，那么那些靠近输出的节点误差不应超过 10%
    # 该方法只衡量 Conv, Gemm 算子的误差情况，如果你对其余算子的误差情况感兴趣，需要手动修改方法逻辑
    # 你可以使用该方法对比不同校准方法对量化结果所产生的影响
    # ------------------------------------------------------------
    reports = graphwise_error_analyse(
        graph=quantized, running_device=DEVICE, collate_fn=collate_fn, dataloader=CALIBRATION
    )



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: Dl

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class TensorShape(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = TensorShape()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsTensorShape(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # TensorShape
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # TensorShape
    def Dim(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from FlatBuffers.Dl.Dimension import Dimension
            obj = Dimension()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # TensorShape
    def DimLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # TensorShape
    def DimIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

def TensorShapeStart(builder):
    builder.StartObject(1)

def Start(builder):
    TensorShapeStart(builder)

def TensorShapeAddDim(builder, dim):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(dim), 0)

def AddDim(builder, dim):
    TensorShapeAddDim(builder, dim)

def TensorShapeStartDimVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartDimVector(builder, numElems):
    return TensorShapeStartDimVector(builder, numElems)

def TensorShapeEnd(builder):
    return builder.EndObject()

def End(builder):
    return TensorShapeEnd(builder)

import FlatBuffers.Dl.Dimension
try:
    from typing import List
except:
    pass

class TensorShapeT(object):

    # TensorShapeT
    def __init__(self):
        self.dim = None  # type: List[FlatBuffers.Dl.Dimension.DimensionT]

    @classmethod
    def InitFromBuf(cls, buf, pos):
        tensorShape = TensorShape()
        tensorShape.Init(buf, pos)
        return cls.InitFromObj(tensorShape)

    @classmethod
    def InitFromPackedBuf(cls, buf, pos=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, pos)
        return cls.InitFromBuf(buf, pos+n)

    @classmethod
    def InitFromObj(cls, tensorShape):
        x = TensorShapeT()
        x._UnPack(tensorShape)
        return x

    # TensorShapeT
    def _UnPack(self, tensorShape):
        if tensorShape is None:
            return
        if not tensorShape.DimIsNone():
            self.dim = []
            for i in range(tensorShape.DimLength()):
                if tensorShape.Dim(i) is None:
                    self.dim.append(None)
                else:
                    dimension_ = FlatBuffers.Dl.Dimension.DimensionT.InitFromObj(tensorShape.Dim(i))
                    self.dim.append(dimension_)

    # TensorShapeT
    def Pack(self, builder):
        if self.dim is not None:
            dimlist = []
            for i in range(len(self.dim)):
                dimlist.append(self.dim[i].Pack(builder))
            TensorShapeStartDimVector(builder, len(self.dim))
            for i in reversed(range(len(self.dim))):
                builder.PrependUOffsetTRelative(dimlist[i])
            dim = builder.EndVector()
        TensorShapeStart(builder)
        if self.dim is not None:
            TensorShapeAddDim(builder, dim)
        tensorShape = TensorShapeEnd(builder)
        return tensorShape



from typing import Iterable, Tuple

import torch

from esp_ppq import BaseGraph, QuantizationSettingFactory, TargetPlatform, convert_any_to_numpy, torch_snr_error
from esp_ppq.api import dispatch_graph, export_ppq_graph, load_onnx_graph, quantize_onnx_model
from esp_ppq.core.data import convert_any_to_torch_tensor
from esp_ppq.executor.torch import TorchExecutor
from esp_ppq.quantization.analyse.graphwise import graphwise_error_analyse

INPUT_SHAPES = {'input.1': [1, 3, 224, 224]}
DEVICE = 'cuda'
QUANT_PLATFORM = TargetPlatform.TRT_INT8
ONNX_PATH = 'model.onnx'
ONNX_OUTPUT_PATH = 'Output/model.onnx'


# ------------------------------------------------------------
# 在这个例子中我们将向你展示如何量化一个 onnx 模型，执行误差分析，并与 onnxruntime 对齐结果
# 在这个例子中，我们特别地为你展示如何量化一个多输入的模型
# 此时你的 Calibration Dataset 应该是一个 list of dictionary
# ------------------------------------------------------------
def generate_calibration_dataset(graph: BaseGraph, num_of_batches: int = 32) -> Tuple[Iterable[dict], torch.Tensor]:
    dataset = []
    for i in range(num_of_batches):
        sample = {name: torch.rand(INPUT_SHAPES[name]) for name in graph.inputs}
        dataset.append(sample)
    return dataset, sample  # last sample


def collate_fn(batch: dict) -> torch.Tensor:
    return {k: v.to(DEVICE) for k, v in batch.items()}


# ------------------------------------------------------------
# 在这里，我们仍然创建一个 QuantizationSetting 对象用来管理量化过程
# 我们将调度方法修改为 conservative，并且要求 PPQ 启动量化微调
# ------------------------------------------------------------
QSetting = QuantizationSettingFactory.default_setting()
QSetting.lsq_optimization = False

# ------------------------------------------------------------
# 准备好 QuantizationSetting 后，我们加载模型，并且要求 ppq 按照规则完成图调度
# ------------------------------------------------------------
graph = load_onnx_graph(onnx_import_file=ONNX_PATH)
graph = dispatch_graph(graph=graph, platform=QUANT_PLATFORM)
for name in graph.inputs:
    if name not in INPUT_SHAPES:
        raise KeyError(f'Graph Input {name} needs a valid shape.')

if len(graph.outputs) != 1:
    raise ValueError('This Script Requires graph to have only 1 output.')

# ------------------------------------------------------------
# 生成校准所需的数据集，我们准备开始完成网络量化任务
# ------------------------------------------------------------
calibration_dataset, sample = generate_calibration_dataset(graph)
quantized = quantize_onnx_model(
    onnx_import_file=ONNX_PATH,
    calib_dataloader=calibration_dataset,
    calib_steps=32,
    input_shape=None,
    inputs=collate_fn(sample),
    setting=QSetting,
    collate_fn=collate_fn,
    platform=QUANT_PLATFORM,
    device=DEVICE,
    verbose=0,
)

# ------------------------------------------------------------
# 在 PPQ 完成网络量化之后，我们特别地保存一下 PPQ 网络执行的结果
# 在本样例的最后，我们将对比 PPQ 与 Onnxruntime 的执行结果是否相同
# ------------------------------------------------------------
executor, reference_outputs = TorchExecutor(quantized), []
for sample in calibration_dataset:
    reference_outputs.append(executor.forward(collate_fn(sample)))

# ------------------------------------------------------------
# 执行网络误差分析，并导出计算图
# ------------------------------------------------------------
graphwise_error_analyse(graph=quantized, running_device=DEVICE, collate_fn=collate_fn, dataloader=calibration_dataset)

export_ppq_graph(graph=quantized, platform=TargetPlatform.ONNXRUNTIME, graph_save_to=ONNX_OUTPUT_PATH)

# -----------------------------------------
# 在最后，我们启动 onnxruntime 并比对结果
# -----------------------------------------
try:
    import onnxruntime
except ImportError as e:
    raise Exception('Onnxruntime is not installed.')

sess = onnxruntime.InferenceSession(ONNX_OUTPUT_PATH, providers=['CUDAExecutionProvider'])
output_name = sess.get_outputs()[0].name

onnxruntime_outputs = []
for sample in calibration_dataset:
    onnxruntime_outputs.append(
        sess.run(output_names=[output_name], input_feed={k: convert_any_to_numpy(v) for k, v in sample.items()})
    )

y_pred, y_real = [], []
for reference_output, onnxruntime_output in zip(reference_outputs, onnxruntime_outputs):
    y_pred.append(convert_any_to_torch_tensor(reference_output[0], device='cpu').unsqueeze(0))
    y_real.append(convert_any_to_torch_tensor(onnxruntime_output[0], device='cpu').unsqueeze(0))
y_pred = torch.cat(y_pred, dim=0)
y_real = torch.cat(y_real, dim=0)
print(f'Simulating Error For {output_name}: {torch_snr_error(y_pred=y_pred, y_real=y_real).item() * 100:.4f}%')



import re

from esp_ppq.api.setting import QuantizationSetting
from esp_ppq.core import (
    OBSERVER_KL_HIST_BINS_MANUL_OVERRIDE,
    PASSIVE_OPERATIONS,
    OperationQuantizationConfig,
    QuantizationPolicy,
    QuantizationProperty,
    QuantizationStates,
    RoundingPolicy,
    TargetPlatform,
    ppq_warning,
)
from esp_ppq.IR import BaseGraph, Operation
from esp_ppq.quantization.optim import QuantizationOptimizationPipeline

from .base import BaseQuantizer


class BaseEspdlQuantizer(BaseQuantizer):
    def __init__(self, graph: BaseGraph) -> None:
        super().__init__(graph=graph)
        self._num_of_bits = 8
        self._quant_min = -128
        self._quant_max = +127
        self._custom_tqc = None

    def build_quant_pipeline(self, setting: QuantizationSetting) -> QuantizationOptimizationPipeline:
        pipeline = super().build_quant_pipeline(setting)
        return pipeline

    def create_espdl_quant_config(
        self,
        operation: Operation,
        num_of_bits: int,
        quant_min: int,
        quant_max: int,
        bias_bits: int,
    ) -> OperationQuantizationConfig:
        base_quant_config = self.create_default_quant_config(
            policy=self.quantize_policy,
            rounding=self.rounding_policy,
            op=operation,
            num_of_bits=num_of_bits,
            exponent_bits=0,
            quant_max=quant_max,
            quant_min=quant_min,
            observer_algorithm="percentile",
        )

        for index in range(operation.num_of_input):
            if not operation.inputs[index].is_parameter:
                base_quant_config.input_quantization_config[index].detail[OBSERVER_KL_HIST_BINS_MANUL_OVERRIDE] = (
                    32 * int(pow(2, num_of_bits - 1))
                )

        for index in range(operation.num_of_output):
            base_quant_config.output_quantization_config[index].detail[OBSERVER_KL_HIST_BINS_MANUL_OVERRIDE] = 32 * int(
                pow(2, num_of_bits - 1)
            )

        if operation.type in {"Conv", "ConvTranspose", "Gemm"}:
            # reset num_of_bits of bias to 32 bits
            assert operation.num_of_input > 0, "Seems you got a Conv layer with no parameters."

            # if operation has bias
            if operation.num_of_input > 2:
                bias_config = base_quant_config.input_quantization_config[-1]
                bias_config.num_of_bits = bias_bits
                bias_config.quant_max = int(pow(2, bias_config.num_of_bits - 1)) - 1
                bias_config.quant_min = -int(pow(2, bias_config.num_of_bits - 1))
                bias_config.state = QuantizationStates.PASSIVE_INIT
                bias_config.observer_algorithm = "minmax"
        elif operation.type in {"GRU", "LSTM"}:
            reset_variables = [3]  # 3:bias for GRU
            if operation.type == "LSTM":
                reset_variables = [3, 6]  # 3:bias, 6: initial_c for LSTM
            for index in reset_variables:
                if operation.num_of_input > index:
                    config = base_quant_config.input_quantization_config[index]
                    config.num_of_bits = 16
                    config.quant_max = int(pow(2, config.num_of_bits - 1)) - 1
                    config.quant_min = -int(pow(2, config.num_of_bits - 1))
                    config.state = QuantizationStates.PASSIVE_INIT
                    config.observer_algorithm = "minmax"

            if operation.num_of_output == 3:
                cell_config = base_quant_config.output_quantization_config[2]
                cell_config.num_of_bits = 16
                cell_config.quant_max = int(pow(2, cell_config.num_of_bits - 1)) - 1
                cell_config.quant_min = -int(pow(2, cell_config.num_of_bits - 1))
                cell_config.state = QuantizationStates.PASSIVE_INIT
                cell_config.observer_algorithm = "minmax"
            for index in range(len(operation.inputs)):
                if (
                    operation.inputs[index].name is None or len(operation.inputs[index].name) == 0
                ):  # Do not quantize bias
                    base_quant_config.input_quantization_config[index].state = QuantizationStates.FP32
        elif operation.type in {"Softmax"}:
            # reset output to float32
            base_quant_config.output_quantization_config[0].state = QuantizationStates.FP32

        if operation.type in PASSIVE_OPERATIONS:
            # Those op are not active op.
            base_quant_config.is_active_quant_op = False

        # Use custom TQC to override configured TQC.
        if self._custom_tqc and self._custom_tqc.get(operation.name):
            configs = self._custom_tqc.get(operation.name)
            for tqc_name in configs.keys():
                if not configs[tqc_name].get("bit_width"):
                    continue

                tqc_index = int(re.findall(r"\d+", tqc_name)[0])
                if "input" in tqc_name:
                    if tqc_index >= operation.num_of_input:
                        ppq_warning(f"Your input tqc index has exceeds num_of_input({operation.num_of_input})!")
                        continue

                    base_quant_config.input_quantization_config[tqc_index].num_of_bits = configs[tqc_name]["bit_width"]
                    base_quant_config.input_quantization_config[tqc_index].quant_max = (
                        +int(pow(2, configs[tqc_name]["bit_width"] - 1)) - 1
                    )
                    base_quant_config.input_quantization_config[tqc_index].quant_min = -int(
                        pow(2, configs[tqc_name]["bit_width"] - 1)
                    )
                    base_quant_config.input_quantization_config[tqc_index].detail[
                        OBSERVER_KL_HIST_BINS_MANUL_OVERRIDE
                    ] = 32 * int(pow(2, configs[tqc_name]["bit_width"] - 1))
                elif "output" in tqc_name:
                    if tqc_index >= operation.num_of_output:
                        ppq_warning(f"Your output tqc index has exceeds num_of_output({operation.num_of_output})!")
                        continue

                    base_quant_config.output_quantization_config[tqc_index].num_of_bits = configs[tqc_name]["bit_width"]
                    base_quant_config.output_quantization_config[tqc_index].quant_max = (
                        +int(pow(2, configs[tqc_name]["bit_width"] - 1)) - 1
                    )
                    base_quant_config.output_quantization_config[tqc_index].quant_min = -int(
                        pow(2, configs[tqc_name]["bit_width"] - 1)
                    )
                    base_quant_config.output_quantization_config[tqc_index].detail[
                        OBSERVER_KL_HIST_BINS_MANUL_OVERRIDE
                    ] = 32 * int(pow(2, configs[tqc_name]["bit_width"] - 1))

        return base_quant_config

    def init_quantize_config(self, operation: Operation) -> OperationQuantizationConfig:
        if operation.platform == self.target_platform:
            num_of_bits = self._num_of_bits
            quant_min = self._quant_min
            quant_max = self._quant_max
        elif operation.platform in [
            TargetPlatform.ESPDL_INT8,
            TargetPlatform.ESPDL_S3_INT8,
            TargetPlatform.ESPDL_C_INT8,
        ]:
            num_of_bits = 8
            quant_min = -128
            quant_max = 127
        elif operation.platform in [
            TargetPlatform.ESPDL_INT16,
            TargetPlatform.ESPDL_H_PRE_INT16,
            TargetPlatform.ESPDL_S3_INT16,
            TargetPlatform.ESPDL_S3_H_PRE_INT16,
            TargetPlatform.ESPDL_C_INT16,
            TargetPlatform.ESPDL_C_H_PRE_INT16,
        ]:
            num_of_bits = 16
            quant_min = -32768
            quant_max = 32767
        else:
            raise KeyError(f"EspdlQuantizer do not support operation platform : {operation.platform}.")

        bias_bits = 32
        if operation.platform == TargetPlatform.ESPDL_S3_INT8:
            bias_bits = 20
        elif operation.platform in [
            TargetPlatform.ESPDL_INT16,
            TargetPlatform.ESPDL_H_PRE_INT16,
            TargetPlatform.ESPDL_S3_INT16,
            TargetPlatform.ESPDL_S3_H_PRE_INT16,
        ]:
            bias_bits = 40
        elif operation.platform in [TargetPlatform.ESPDL_C_INT16, TargetPlatform.ESPDL_C_H_PRE_INT16]:
            bias_bits = 64

        return self.create_espdl_quant_config(operation, num_of_bits, quant_min, quant_max, bias_bits)

    @property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.ESPDL_INT8

    @property
    def default_platform(self) -> TargetPlatform:
        return TargetPlatform.FP32

    @property
    def quant_operation_types(self) -> set:
        return {
            "Conv",
            "ConvTranspose",
            "Gemm",
            "GRU",
            "LSTM",
            "Relu",
            "PRelu",
            "Clip",
            "Pad",
            "Resize",
            "MaxPool",
            "AveragePool",
            "GlobalMaxPool",
            "GlobalAveragePool",
            "Softmax",
            "Mul",
            "Add",
            "Max",
            "Sub",
            "Div",
            "Neg",
            "Reshape",
            "LeakyRelu",
            "Concat",
            "Sigmoid",
            "Interp",
            "ReduceL1",
            "ReduceL2",
            "ReduceMean",
            "ReduceMin",
            "ReduceMax",
            "ReduceProd",
            "ReduceSum",
            "ReduceSumSquare",
            "ReduceLogSum",
            "ReduceLogSumExp",
            "Transpose",
            "Slice",
            "Flatten",
            "HardSwish",
            "HardSigmoid",
            "MatMul",
            "Attention",
            "LayerNormalization",
            "Gelu",
            "PPQBiasFusedMatMul",
            "Split",
            "Gather",
            "ScatterND",
            "Tanh",
            "Elu",
            "Greater",
            "Less",
            "Equal",
            "GreaterOrEqual",
            "LessOrEqual",
            "ReverseSequence",
            "Identity",
            "Swish",
            'Squeeze',
            'Unsqueeze',
            'Exp',
            'DepthToSpace',
            'SpaceToDepth',
            'InsertZeros',
        }

    @property
    def quantize_policy(self) -> QuantizationPolicy:
        return QuantizationPolicy(
            QuantizationProperty.SYMMETRICAL
            + QuantizationProperty.LINEAR
            + QuantizationProperty.PER_TENSOR
            + QuantizationProperty.POWER_OF_2
        )

    @property
    def rounding_policy(self):
        return RoundingPolicy.ROUND_HALF_EVEN

    @property
    def activation_fusion_types(self) -> set:
        """
        我不知道这个对不对, 这个是遵循 Xilinx FPGA 的修改，
        如果你的网络中有特殊的激活函数，我建议你手动调整这个融合选项

        Returns:
            set: _description_
        """
        return {"Relu", "Clip"}

    @property
    def custom_tqc(self) -> dict:
        return self._custom_tqc

    # The custom_op_tqc format is as follows:
    # {
    #     'op_name': {
    #         'input_0': {
    #             'bit_width': 8
    #             ......
    #         }
    #         ......
    #         'output_0': {
    #             'bit_width': 8
    #             ......
    #         }
    #     }
    # }
    @custom_tqc.setter
    def custom_tqc(self, custom_op_tqc: dict):
        self._custom_tqc = custom_op_tqc


class EspdlQuantizer(BaseEspdlQuantizer):
    def __init__(self, graph: BaseGraph) -> None:
        super().__init__(graph=graph)


class EspdlInt16Quantizer(BaseEspdlQuantizer):
    def __init__(self, graph: BaseGraph) -> None:
        super().__init__(graph=graph)
        self._num_of_bits = 16
        self._quant_min = -32768
        self._quant_max = +32767
        self._custom_tqc = None

    @property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.ESPDL_INT16


class EspdlHPreInt16Quantizer(EspdlInt16Quantizer):
    def __init__(self, graph: BaseGraph) -> None:
        super().__init__(graph=graph)

    @property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.ESPDL_H_PRE_INT16


class EspdlS3Quantizer(BaseEspdlQuantizer):
    def __init__(self, graph: BaseGraph) -> None:
        super().__init__(graph=graph)

    @property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.ESPDL_S3_INT8

    @property
    def rounding_policy(self):
        return RoundingPolicy.ROUND_HALF_UP


class EspdlS3Int16Quantizer(BaseEspdlQuantizer):
    def __init__(self, graph: BaseGraph) -> None:
        super().__init__(graph=graph)
        self._num_of_bits = 16
        self._quant_min = -32768
        self._quant_max = +32767
        self._custom_tqc = None

    @property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.ESPDL_S3_INT16

    @property
    def rounding_policy(self):
        return RoundingPolicy.ROUND_HALF_UP


class EspdlS3HPreInt16Quantizer(EspdlS3Int16Quantizer):
    def __init__(self, graph: BaseGraph) -> None:
        super().__init__(graph=graph)

    @property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.ESPDL_S3_H_PRE_INT16


class EspdlCQuantizer(BaseEspdlQuantizer):
    def __init__(self, graph: BaseGraph) -> None:
        super().__init__(graph=graph)

    @property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.ESPDL_C_INT8

    @property
    def rounding_policy(self):
        return RoundingPolicy.ROUND_HALF_UP


class EspdlCInt16Quantizer(BaseEspdlQuantizer):
    def __init__(self, graph: BaseGraph) -> None:
        super().__init__(graph=graph)
        self._num_of_bits = 16
        self._quant_min = -32768
        self._quant_max = +32767
        self._custom_tqc = None

    @property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.ESPDL_C_INT16

    @property
    def rounding_policy(self):
        return RoundingPolicy.ROUND_HALF_UP


class EspdlCHPreInt16Quantizer(EspdlCInt16Quantizer):
    def __init__(self, graph: BaseGraph) -> None:
        super().__init__(graph=graph)

    @property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.ESPDL_C_H_PRE_INT16



import os
import sys
from typing import List

import numpy as np
import torch

from esp_ppq.core import (
    DataType,
    OperationQuantizationConfig,
    QuantizationProperty,
    QuantizationStates,
    QuantizationVisibility,
    TensorQuantizationConfig,
    convert_any_to_numpy,
    convert_any_to_torch_tensor,
)
from esp_ppq.IR import BaseGraph, Operation, OperationExporter, Variable
from esp_ppq.IR.quantize import QuantableOperation
from esp_ppq.log import NaiveLogger
from esp_ppq.parser.espdl.espdl_typedef import EspQuantType, ExporterPatternInfo
from esp_ppq.utils.round import ppq_tensor_round

logger = NaiveLogger.get_logger("ESPDL")


def transpose_shape(input_shape, perm: List[int]) -> List[int]:
    if not perm:
        return input_shape
    return [input_shape[i] for i in perm]


def get_inverse_transpose(perm: List[int]) -> List[int]:
    """
    tensor == inverse_transpose(transpose(tensor))
    """
    # return perm
    return [perm.index(i) for i in range(len(perm))]


def get_default_perm(var: Variable) -> List[int]:
    """
    return the default permute for given variable, [0,1,2,3,...]
    """
    if not var or not var.shape:
        return []

    return [i for i in range(len(var.shape))]


def insert_transpose_node(graph: BaseGraph, var: Variable, op: Operation, perm: List[int]) -> Operation:
    """
    Insert a Transpose Node on given variable, according to given perm.
    """
    created = None
    if op and perm != range(len(perm)):
        logger.debug(f"insert transpose node: op: {op.name}, var:{var.name}, perm:{perm}")
        if var in op.inputs:
            created = graph.create_operation(op_type="Transpose", attributes={"perm": perm})
            var_index = op.inputs.index(var)
            if isinstance(op, QuantableOperation):
                # For transpose op,  input_quantization_config == output_quantization_config
                new_config = OperationQuantizationConfig(
                    [op.input_quant_config[var_index]],
                    [op.input_quant_config[var_index]],
                )
                created = QuantableOperation(created, new_config, op.platform)
                graph.operations[created.name] = created

            graph.insert_op_before(A=created, B=op, input_idx=var_index)
            new_var = created.outputs[0]
            new_var.shape = [var.shape[i] for i in perm]
            new_var.is_parameter = False
            new_var.dtype = var.dtype

            info = ExporterPatternInfo()
            info.add_var_permute(created.outputs[0].name, get_default_perm(created.outputs[0]))

        else:
            raise ValueError(f"Unexpected Error in Exporting Op {op.name}({op.type}).")

    return created


def restore_origin_shape(op: Operation, graph: BaseGraph):
    info = ExporterPatternInfo()
    for var in op.inputs:
        if var.is_parameter:
            continue

        var_perm = info.get_var_permute(var.name)
        if var_perm and var_perm != get_default_perm(var):
            # There is already a permute, but this op need keep origin shape
            # A transpose node needs to be inserted into the word.
            inverse_perm = get_inverse_transpose(var_perm)
            insert_transpose_node(graph, var, op, inverse_perm)
        else:
            info.add_var_permute(var.name, get_default_perm(var))

    for var in op.outputs:
        info.add_var_permute(var.name, get_default_perm(var))
    return op


def insert_concat_node(
    graph: BaseGraph,
    insert_op_var: Variable,
    insert_op: Operation,
    link_vars: List[Variable],
    link_vars_src_op: List[Operation],
    axis: int = 0,
) -> Operation:
    """
    Insert a Concat Node on given insert_op_var, according to given axis. And using link_vars as the other input,
    these function will use the TQC of the first input as the TQC for all the inputs and output.

    """
    created = None
    if insert_op and insert_op_var in insert_op.inputs:
        logger.debug(
            f"insert concat node: insert_op: {insert_op.name}, insert_op_var:{insert_op_var.name}, link_vars number:{len(link_vars)}"
        )
        if axis < 0:
            axis = axis + len(insert_op_var.shape)
        created = graph.create_operation(op_type="Concat", attributes={"axis": axis})
        var_index = insert_op.inputs.index(insert_op_var)
        if isinstance(insert_op, QuantableOperation):
            # For concat insert_op,  input_quantization_config == output_quantization_config
            input_config = insert_op.input_quant_config[var_index].copy()
            input_config.state = QuantizationStates.PASSIVE
            new_config = OperationQuantizationConfig(
                [input_config] * (len(link_vars) + 1),
                [insert_op.input_quant_config[var_index]],
            )
            created = QuantableOperation(created, new_config, insert_op.platform)
            graph.operations[created.name] = created

        output_shape = insert_op_var.shape.copy()
        for var in link_vars:
            assert len(var.shape) == len(insert_op_var.shape), (
                f"The inputs of concat have inconsistent numbers of dimensions {len(insert_op_var.shape)}, {len(var.shape)}."
            )
            output_shape[axis] = output_shape[axis] + var.shape[axis]

        graph.insert_op_before(A=created, B=insert_op, input_idx=var_index)
        new_var = created.outputs[0]
        new_var.shape = output_shape
        new_var.is_parameter = False
        new_var.dtype = insert_op_var.dtype

        for var, src_op in zip(link_vars, link_vars_src_op):
            graph.create_link_with_op(A=src_op, B=created, variable=var)

    else:
        raise ValueError(f"Unexpected Error in Exporting Op {insert_op.name}({insert_op.type}).")

    return created


def insert_slice_node(
    graph: BaseGraph,
    var: Variable,
    op: Operation,
    starts: List[int],
    ends: List[int],
    axes: List[int],
    steps: List[int],
) -> Operation:
    """
    Insert a Slice Node on given variable, according to given starts, ends, axes, steps.
    """
    created = None
    if op and var in op.inputs:
        logger.debug(
            f"insert slice node: op: {op.name}, var: {var.name}, starts: {starts}, ends: {ends}, axes: {axes}, steps: {steps}"
        )
        created = graph.create_operation(op_type="Slice")

        var_index = op.inputs.index(var)
        if isinstance(op, QuantableOperation):
            # For slice op,  input_quantization_config[0] == output_quantization_config
            input0_config = op.input_quant_config[var_index].copy()
            input0_config.state = QuantizationStates.OVERLAPPED

            starts_config = op.input_quant_config[var_index].copy()
            starts_config.state = QuantizationStates.FP32
            starts_config.observer_algorithm = "percentile"
            ends_config = starts_config.copy()
            axes_config = starts_config.copy()
            steps_config = starts_config.copy()

            output0_config = op.input_quant_config[var_index].copy()
            output0_config.state = QuantizationStates.OVERLAPPED
            output0_config.dominated_by = input0_config

            new_config = OperationQuantizationConfig(
                [input0_config, starts_config, ends_config, axes_config, steps_config],
                [output0_config],
            )
            created = QuantableOperation(created, new_config, op.platform)
            graph.operations[created.name] = created

        dim = len(var.shape)
        output_shape = [0] * dim
        for i in range(len(starts)):
            axis = i
            if axes:
                axis = (axes[i] + dim) % dim

            start_i = starts[i] + var.shape[axis] if starts[i] < 0 else starts[i] % (var.shape[axis] + 1)
            end_i = ends[i] + var.shape[axis] if ends[i] < 0 else ends[i] % (var.shape[axis] + 1)
            if start_i >= end_i:
                raise ValueError(f"Unexpected value, start_i: {start_i}, end_i: {end_i}.")
            else:
                if steps:
                    output_shape[axis] = 1 + (end_i - start_i - 1) // steps[i]
                else:
                    output_shape[axis] = end_i - start_i

        graph.insert_op_before(A=created, B=op, input_idx=var_index)
        new_var = created.outputs[0]
        new_var.shape = output_shape
        new_var.is_parameter = var.is_parameter
        new_var.dtype = var.dtype

        starts_param = graph.create_variable(
            value=torch.Tensor(starts).to(torch.int64), is_parameter=True, dest_ops=[created]
        )
        starts_param.dtype = DataType.INT64
        ends_param = graph.create_variable(
            value=torch.Tensor(ends).to(torch.int64), is_parameter=True, dest_ops=[created]
        )
        ends_param.dtype = DataType.INT64
        axes_param = graph.create_variable(
            value=torch.Tensor(axes).to(torch.int64), is_parameter=True, dest_ops=[created]
        )
        axes_param.dtype = DataType.INT64
        steps_param = graph.create_variable(
            value=torch.Tensor(steps).to(torch.int64), is_parameter=True, dest_ops=[created]
        )
        steps_param.dtype = DataType.INT64

    else:
        raise ValueError(f"Unexpected Error in insert slice node, var: {var.name}, Op: {op.name}.")

    return created


def insert_reshape_node(
    graph: BaseGraph, var: Variable, op: Operation, shape: List[int], allowzero: int = 0
) -> Operation:
    """
    Insert a Reshape Node on given variable, according to given shape.
    """
    created = None
    if op and var in op.inputs:
        logger.debug(f"insert reshape node: op: {op.name}, var: {var.name}, shape: {shape}")
        created = graph.create_operation(op_type="Reshape", attributes={"allowzero": allowzero})

        var_index = op.inputs.index(var)
        if isinstance(op, QuantableOperation):
            # For reshape op,  input_quantization_config[0] == output_quantization_config
            shape_config = op.input_quant_config[var_index].copy()
            shape_config.state = QuantizationStates.FP32
            shape_config.observer_algorithm = "percentile"

            new_config = OperationQuantizationConfig(
                [op.input_quant_config[var_index], shape_config],
                [op.input_quant_config[var_index]],
            )
            created = QuantableOperation(created, new_config, op.platform)
            graph.operations[created.name] = created

        graph.insert_op_before(A=created, B=op, input_idx=var_index)
        new_var = created.outputs[0]
        new_var.shape = shape
        new_var.is_parameter = var.is_parameter
        new_var.dtype = var.dtype

        shape_param = graph.create_variable(
            value=torch.Tensor(shape).to(torch.int64), is_parameter=True, dest_ops=[created]
        )
        shape_param.dtype = DataType.INT64

    else:
        raise ValueError(f"Unexpected Error in insert reshape node, var: {var.name}, Op: {op.name}.")

    return created


def fuse_downstream_operation(
    graph: BaseGraph,
    fusing_downstream_op: Operation,
    keep_coherence: bool = False,
    remove_unlinked_variable: bool = False,
):
    """Remove operation from graph, this function will unlink removing
    operation from current graph, pop it from graph.operations, and remove
    it from all its input and output variables.

    Parameters of this removing operations will be removed from graph by this function, without warning.

    Args:
        fusing_downstream_op (Operation): [description]

        keep_coherence (bool): if keep_coherence = True,
            PPQ will link downstream operations of removing op to the upstream operation.
            if there is more than 1 input and output variable, ppq will link input[0] with output[0]
    """
    if fusing_downstream_op.name not in graph.operations:
        raise KeyError(f"Can not remove operation {fusing_downstream_op.name}, operation not found.")

    # removing all parameters first.
    for parameter in fusing_downstream_op.inputs.copy():
        if keep_coherence and fusing_downstream_op.type in {"Constant", "Identity"}:
            break
        if parameter.is_parameter:
            parameter.dest_ops.clear()
            parameter.value = None  # clear memory.
            fusing_downstream_op.inputs.remove(parameter)

            graph.variables.pop(parameter.name)

    related_vars = [var for var in fusing_downstream_op.inputs + fusing_downstream_op.outputs]
    input_var, output_var = (
        fusing_downstream_op.inputs[0] if fusing_downstream_op.num_of_input >= 1 else None,
        fusing_downstream_op.outputs[0] if fusing_downstream_op.num_of_output >= 1 else None,
    )

    # remove operation from its output variables
    for _output_var in fusing_downstream_op.outputs:
        _output_var.source_op = None
    fusing_downstream_op.outputs.clear()

    # remove operation from its input variables
    for _input_var in fusing_downstream_op.inputs:
        if fusing_downstream_op in _input_var.dest_ops:
            _input_var.dest_ops.remove(fusing_downstream_op)
    fusing_downstream_op.inputs.clear()

    if input_var is not None and output_var is not None and keep_coherence:
        removing_var = input_var
        source_op = removing_var.source_op
        source_op.outputs[source_op.outputs.index(removing_var)] = output_var
        output_var.source_op = source_op
        removing_var.source_op = None
        removing_var.dest_ops.clear()
        graph.remove_variable(removing_var)

    graph.operations.pop(fusing_downstream_op.name)

    if remove_unlinked_variable:
        for var in related_vars:
            if var.source_op is None and len(var.dest_ops) == 0 and var.name in graph.variables:
                graph.remove_variable(var)

    return graph


def infer_qtype(config: TensorQuantizationConfig):
    offset_dtype, value_dtype = torch.int8, torch.int8
    if config.policy.has_property(QuantizationProperty.ASYMMETRICAL):
        offset_dtype = torch.uint8
        value_dtype = torch.uint8
    if config.num_of_bits > 8:
        offset_dtype = torch.int16
        value_dtype = torch.int16
    elif config.num_of_bits > 16:
        offset_dtype = torch.int32
        value_dtype = torch.int32
    return offset_dtype, value_dtype


def insert_quantize_node(graph: BaseGraph, var: Variable, config: TensorQuantizationConfig, op: Operation) -> Operation:
    """
    Insert a Quantize Node on given variable, according to given TensorQuantizationConfig.
    """
    if config.policy.has_property(QuantizationProperty.LINEAR):
        # Following code will export Linear Quantization Config
        # That is for FP32 -> INT
        offset_dtype, value_type = infer_qtype(config)
        scale = convert_any_to_torch_tensor(config.scale.clone(), dtype=torch.float32)
        offset = ppq_tensor_round(config.offset.clone()).type(offset_dtype)

        created = graph.create_operation(op_type="QuantizeLinear", attributes={})
        if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
            created.attributes["axis"] = config.channel_axis
        else:
            created.attributes["axis"] = None

        if var in op.inputs:
            graph.insert_op_before(A=created, B=op, input_idx=op.inputs.index(var))
        elif var in op.outputs:
            graph.insert_op_after(A=created, B=op, output_idx=op.outputs.index(var))
        else:
            raise ValueError(f"Unexpected Error in Exporting Op {op.name}({op.type}).")
        created.platform = op.platform

        graph.create_variable(name=None, value=scale, is_parameter=True, dest_ops=[created])
        graph.create_variable(name=None, value=offset, is_parameter=True, dest_ops=[created])

        created.outputs[0].dtype = value_type
        created.outputs[0].shape = var.shape
        created.inputs[0].shape = var.shape
        return created

    else:
        raise TypeError(
            f"PPQ Can not export quantization information with variable {var.name}, Unexpected Quantization property."
        )


def insert_requantize_node(
    graph: BaseGraph,
    var: Variable,
    upstream_config: TensorQuantizationConfig,
    config: TensorQuantizationConfig,
    op: Operation,
) -> Operation:
    """
    Insert a ReQuantize Node on given variable, according to given TensorQuantizationConfig.
    """
    if config.policy.has_property(QuantizationProperty.LINEAR):
        upstream_offset_dtype, upstream_value_type = infer_qtype(upstream_config)
        upstream_scale = convert_any_to_torch_tensor(upstream_config.scale.clone(), dtype=torch.float32)
        upstream_offset = ppq_tensor_round(upstream_config.offset.clone()).type(torch.float)
        offset_dtype, value_type = infer_qtype(config)
        scale = convert_any_to_torch_tensor(config.scale.clone(), dtype=torch.float32)
        offset = ppq_tensor_round(config.offset.clone()).type(torch.float)

        created = graph.create_operation(op_type="RequantizeLinear", attributes={})
        if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
            created.attributes["axis"] = config.channel_axis
        else:
            created.attributes["axis"] = None

        if var in op.inputs:
            graph.insert_op_before(A=created, B=op, input_idx=op.inputs.index(var))
        elif var in op.outputs:
            graph.insert_op_after(A=created, B=op, output_idx=op.outputs.index(var))
        else:
            raise ValueError(f"Unexpected Error in Exporting Op {op.name}({op.type}).")
        created.platform = op.platform
        rescale = scale / upstream_scale
        reoffset = ppq_tensor_round(offset - ppq_tensor_round(upstream_offset / rescale, config.rounding)).type(
            offset_dtype
        )

        graph.create_variable(name=None, value=rescale, is_parameter=True, dest_ops=[created])
        graph.create_variable(name=None, value=reoffset, is_parameter=True, dest_ops=[created])

        created.inputs[0].dtype = upstream_value_type
        created.inputs[0].shape = var.shape
        created.outputs[0].shape = var.shape
        created.outputs[0].dtype = value_type
        return created

    else:
        raise TypeError(
            f"PPQ Can not export quantization information with variable {var.name}, Unexpected Quantization property."
        )


def insert_dequantize_node(
    graph: BaseGraph, var: Variable, config: TensorQuantizationConfig, op: Operation
) -> Operation:
    """
    Insert a DeQuantize Node on given variable, according to given TensorQuantizationConfig.
    """
    if config.policy.has_property(QuantizationProperty.LINEAR):
        offset_dtype, value_type = infer_qtype(config)
        scale = convert_any_to_torch_tensor(config.scale.clone(), dtype=torch.float32)
        offset = ppq_tensor_round(config.offset.clone()).type(offset_dtype)

        created = graph.create_operation(op_type="DequantizeLinear", attributes={})
        if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
            created.attributes["axis"] = config.channel_axis
        else:
            created.attributes["axis"] = None

        if var in op.inputs:
            graph.insert_op_before(A=created, B=op, input_idx=op.inputs.index(var))
        elif var in op.outputs:
            graph.insert_op_after(A=created, B=op, output_idx=op.outputs.index(var))
        else:
            raise ValueError(f"Unexpected Error in Exporting Op {op.name}({op.type}).")

        graph.create_variable(name=None, value=scale, is_parameter=True, dest_ops=[created])
        graph.create_variable(name=None, value=offset, is_parameter=True, dest_ops=[created])

        created.inputs[0].dtype = value_type
        created.inputs[0].shape = var.shape
        created.outputs[0].shape = var.shape
        created.outputs[0].dtype = torch.float32
        return created

    else:
        raise TypeError(
            f"PPQ Can not export quantization information with variable {var.name}, Unexpected Quantization property."
        )



from typing import Iterable

from esp_ppq.executor import BaseGraphExecutor
from esp_ppq.IR.base.graph import BaseGraph

from .base import QuantizationOptimizationPass


class ExtensionPass(QuantizationOptimizationPass):
    """ExtensionPass 并没有什么用，它就是告诉你你可以像这样写一个 pass。 你可以直接改写 ExtensionPass
    的逻辑来实现你的功能，并将修改后的代码提交到 github.

    不过比较我们已经为 ExtensionPass 创建了一个 TemplateSetting 用来给它传递参数
        你可以去 esp_ppq.api.setting.py 里面找到它

    There is nothing in ExtensionPass, it is literally an empty pass,
        -- just show you how to create your own pass.

    A TemplateSetting class has been created for passing parameter to this pass.
        You can find it in esp_ppq.api.setting.py

    You can overwrite logic inside this pass.
    """

    def __init__(self, parameter: str) -> None:
        self.parameter = parameter
        super().__init__(name='PPQ Extension Pass')

    def optimize(self, graph: BaseGraph, dataloader: Iterable, executor: BaseGraphExecutor, **kwargs) -> None:
        assert isinstance(graph, BaseGraph)

        print('You are invoking Extension Pass now.')



from .allin import AllinDispatcher
from .base import GraphDispatcher, reverse_tracing_pattern, value_tracing_pattern
from .dispatchers import AggresiveDispatcher, ConservativeDispatcher, PointDispatcher, PPLNNDispatcher
from .perseus import Perseus

# Do not forget register your dispather here.

DISPATCHER_TABLE = {
    "conservative": ConservativeDispatcher,
    "pplnn": PPLNNDispatcher,
    "aggresive": AggresiveDispatcher,
    "pointwise": PointDispatcher,
    "allin": AllinDispatcher,
    'perseus': Perseus,
}



# Test Quantization System Performace on Image Classification Models with ILSVRC2012 Dataset
#
#   1. How to use:
#      Run this script with python directly.

# Quantizer Configuration
SYMMETRICAL = True
POWER_OF_2 = True
PER_CHANNEL = False
FP8 = True
BIT_WIDTH = 8

# Should contains model file(.onnx)
MODEL_DIR = 'QuantZoo/Model/Imagenet'

# Should contains Calib & Test Img Folder
CALIB_DIR = 'QuantZoo/Data/Imagenet/Calib'
TEST_DIR = 'QuantZoo/Data/Imagenet/Test'

# calibration & test batchsize
BATCHSIZE = 32

# write report to here
REPORT_DIR = 'QuantZoo/Reports'

CONFIGS = [
    {'Model': 'repvgg', 'Output': ['input.172']},
    {'Model': 'efficientnet_v1_b0', 'Output': ['/features/features.8/features.8.2/Mul_output_0']},
    {'Model': 'efficientnet_v1_b1', 'Output': ['/features/features.8/features.8.2/Mul_output_0']},
    {'Model': 'efficientnet_v2_s', 'Output': ['/features/features.7/features.7.2/Mul_output_0']},
    {'Model': 'mnasnet0_5', 'Output': ['/layers/layers.16/Relu_output_0']},
    {'Model': 'mnasnet1_0', 'Output': ['/layers/layers.16/Relu_output_0']},
    {'Model': 'mobilenet_v2', 'Output': ['/features/features.18/features.18.2/Clip_output_0']},
    {'Model': 'resnet18', 'Output': ['/layer4/layer4.1/relu_1/Relu_output_0']},
    {'Model': 'resnet50', 'Output': ['/layer4/layer4.2/relu_2/Relu_output_0']},
    {'Model': 'mobilenet_v3_large', 'Output': ['/classifier/classifier.1/Mul_output_0']},
    {'Model': 'mobilenet_v3_small', 'Output': ['/classifier/classifier.1/Mul_output_0']},
    {'Model': 'v100_gpu64@5ms_top1@71.6_finetune@25', 'Output': ['471']},
    {'Model': 'v100_gpu64@6ms_top1@73.0_finetune@25', 'Output': ['471']},
    {'Model': 'lcnet_050', 'Output': ['/act2/Mul_output_0']},
    {'Model': 'lcnet_100', 'Output': ['/act2/Mul_output_0']},
    {
        # vit_b_16 requires BATCHSIZE = 1!
        'Model': 'vit_b_16',
        'Output': ['onnx::Gather_1703'],
    },
]

import os

import torch
from QuantZoo.Data.Imagenet.Eval import evaluate_ppq_module_with_imagenet, load_imagenet_from_directory
from QuantZoo.Quantizers import MyFP8Quantizer, MyInt8Quantizer
from QuantZoo.Util import error_analyze

import esp_ppq.lib as PFL
from esp_ppq.api import ENABLE_CUDA_KERNEL, load_onnx_graph
from esp_ppq.core import TargetPlatform
from esp_ppq.executor import TorchExecutor
from esp_ppq.quantization.optim import (
    LayerwiseEqualizationPass,
    LearnedStepSizePass,
    ParameterQuantizePass,
    RuntimeCalibrationPass,
)

calib_loader = load_imagenet_from_directory(
    directory=CALIB_DIR, batchsize=BATCHSIZE, shuffle=False, require_label=False, num_of_workers=8
)


test_loader = load_imagenet_from_directory(
    directory=TEST_DIR, batchsize=BATCHSIZE, shuffle=False, require_label=True, num_of_workers=8
)


with ENABLE_CUDA_KERNEL():
    for config in CONFIGS:
        model = config['Model']
        monitoring_vars = config['Output']

        print(f"Ready to run quant benchmark on {model}")
        graph = load_onnx_graph(onnx_import_file=os.path.join(MODEL_DIR, model + '.onnx'))

        if model == 'vit_b_16':
            if BATCHSIZE == 32:
                raise Exception('To Evaluate vit_b_16, change batchsize to 1, change calibration method to minmax.')
            from esp_ppq.IR import GraphMerger

            processor = GraphMerger(graph)
            processor.fuse_matmul_add()
            processor.fuse_layernorm()
            processor.fuse_gelu()

        quantizer = MyInt8Quantizer(
            graph=graph, sym=SYMMETRICAL, power_of_2=POWER_OF_2, num_of_bits=BIT_WIDTH, per_channel=PER_CHANNEL
        )
        if FP8:
            quantizer = MyFP8Quantizer(graph=graph, calibration='floating')

        # convert op to quantable-op
        for name, op in graph.operations.items():
            if op.type in {'Conv', 'ConvTranspose', 'MatMul', 'Gemm', 'PPQBiasFusedMatMul', 'LayerNormalization'}:
                quantizer.quantize_operation(name, platform=TargetPlatform.INT8)

        # build quant pipeline.
        pipeline = PFL.Pipeline(
            [
                # LayerwiseEqualizationPass(iteration=10),
                ParameterQuantizePass(),
                RuntimeCalibrationPass(),
                # LearnedStepSizePass(steps=500, collecting_device='cuda', block_size=5)
            ]
        )

        # call pipeline.
        executor = TorchExecutor(graph=graph)
        executor.tracing_operation_meta(torch.zeros(size=[BATCHSIZE, 3, 224, 224]).cuda())

        pipeline.optimize(
            graph=graph,
            dataloader=calib_loader,
            verbose=True,
            calib_steps=32,
            collate_fn=lambda x: x.to('cuda'),
            executor=executor,
        )

        # evaluation
        acc = evaluate_ppq_module_with_imagenet(
            model=graph, imagenet_validation_loader=test_loader, batchsize=BATCHSIZE, device='cuda', verbose=False
        )
        print(f'Model Classify Accurarcy = {acc: .4f}%')

        # error analyze
        performance = error_analyze(
            graph=graph,
            outputs=monitoring_vars,
            dataloader=test_loader,
            collate_fn=lambda x: x[0].to('cuda'),
            verbose=True,
        )



from abc import ABCMeta, abstractmethod
from collections import defaultdict, deque
from typing import Any, Callable, Dict, Iterable, Iterator, List, Set, Union

from esp_ppq.core import ppq_warning

from .base.command import GraphCommand, GraphCommandType
from .base.graph import BaseGraph, Operation
from .processer import GraphCommandProcessor


class PointPattern(Callable, metaclass=ABCMeta):
    @abstractmethod
    def match(self, operation: Operation) -> bool:
        pass

    def __call__(self, operation: Operation) -> bool:
        return self.match(operation)


class RelayPattern(Callable, metaclass=ABCMeta):
    @abstractmethod
    def match(self, from_where: Operation, to_where: Operation) -> bool:
        pass

    def __call__(self, from_where: Operation, to_where: Operation) -> bool:
        return self.match(from_where, to_where)


class Path(Iterable):
    def __init__(self, operation: Operation = None) -> None:
        self._container = deque()
        if operation is not None:
            self._container.append(operation)

    def append(self, op: Operation):
        self._container.append(op)
        return self

    def append_left(self, op: Operation):
        self._container.appendleft(op)
        return self

    def __iter__(self) -> Iterator[Operation]:
        return self._container.__iter__()

    def __getitem__(self, index: int) -> Operation:
        return self._container[index]

    def __str__(self) -> str:
        return ''.join([str(_) for _ in self._container.__str__()])

    def tolist(self) -> List[Operation]:
        return list(self._container)

    def copy(self):
        c = Path()
        c._container = self._container.copy()
        return c


class OperationSet(set):
    def __init__(self) -> None:
        super().__init__()

    def add(self, element: Operation):
        if not isinstance(element, Operation):
            raise TypeError('Operation Set can only contains operation instance.')
        super().add(element)
        return self

    def __iter__(self) -> Iterator[Operation]:
        return super().__iter__()

    def filter(self, condition: PointPattern):
        removing = []
        for item in self:
            if condition(item):
                removing.append(item)
        for item in removing:
            self.remove(item)


class TraversalCommand(GraphCommand):
    def __init__(
        self,
        sp_expr: Union[PointPattern, Callable],
        rp_expr: Union[RelayPattern, Callable],
        ep_expr: Union[PointPattern, Callable],
        direction: str = 'down',
        matching_type: str = 'path',
    ) -> None:
        """
            TraversalCommand 是一个用于表示图检索指令的结构体
            一个最简单的例子：
                sp_expr = lamdba x: x.type == 'Conv'
                rp_expr = lamdba x: x.type == 'Relu'
                ep_expr = lamdba x: x.type == 'Conv'
                limitation = None
                direction = 'down'
            该指令检索出从任意 Conv 出发，到任意 Conv 的所有可行路径
            其中路径上含有任意多个 Relu 节点，并且只能包含 Relu

            TraversalCommand is a graph search class
            Example:
                sp_expr = lamdba x: x.type == 'Conv'
                rp_expr = lamdba x: x.type == 'Relu'
                ep_expr = lamdba x: x.type == 'Conv'
                limitation = None
                direction = 'down'
            this Command searches paths which starts with any
            Conv op, visits any number of Relu ops, and finally
            ends with any Conv op
        Args:
            sp_expr (Union[PointPattern, Callable]):
                start point expression, 用于匹配检索起点的表达式
            rp_expr (Union[RelayPattern, Callable]):
                relay point expression, 用于匹配检索中继点的表达式
            ep_expr (Union[PointPattern, Callable]):
                end point expression, 用于匹配检索终点的表达式
            direction (str, optional):
                图检索方向，up, down. Defaults to 'down'.
            matching_type (str, optional)
                指定匹配模式，可以指定为 path，则系统匹配完整路径
                指定为 opset，则系统只匹配涉及到的节点
        """
        self._direction = direction
        self._sp_expr = sp_expr
        self._rp_expr = rp_expr
        self._ep_expr = ep_expr
        if matching_type == 'path':
            super().__init__(GraphCommandType.TRAVERSAL_PATTERN_MATCHING)
        elif matching_type == 'opset':
            super().__init__(GraphCommandType.TRAVERSAL_OPSET_MATCHING)
        else:
            raise ValueError('PPQ only support "opset" matching and "path" matching for now.')

    @staticmethod
    def complie(query: str):
        """compile 函数把一个查询字符串编译成一个 TraversalCommand。 我们还没有具体实现这个函数，但我们已经定义好了语法：

        查询字符串应该是下面的格式：
        SELECT  ["START"|"END"|"PATH"]
        FROM    [START PATTERN CLAUSE]
        TO      [END PATTERN CLAUSE]
        THROUGH [RELAY PATTERN CLAUSE]
        WHERE   [PATH PATTERN CLAUSE]

        例如:
        SELECT  "START"
        FROM    Conv, Gemm
        TO      Relu
        THROUGH Any
        WHERE   Path.length < 3
        语句将返回从 Conv, Gemm 出发，到 Relu 的所有可能路径
        从中筛选出路径长度小于 3 的，并且将其中的所有起点作为集合并返回

        SELECT  "PATH"
        FROM    Conv
        TO      Relu, Clip
        THROUGH Any
        WHERE   Path.length < 2
        语句将返回从 Conv 出发，到 Relu, Clip 的所有可能路径
        从中筛选出路径长度小于 2 的，并且将路径本身作为结果返回
        """
        pass


class GraphPattern:
    def __init__(self, node_patterns: List[Callable], edges: List[List[int]]) -> None:
        """Pattern Tree 是一个用来表示图模式的结构体 这将在图中检索任意一个子图.

        你将使用 Graph Pattern 定义你的子图结构
        使用 patterns 确定每一个节点需要满足的条件
        使用 edges 将节点们彼此相连从而构成图结构

        构成的图必须可以进行拓扑排序，不可以检索有环结构，不可以检索不连通结构
        例子：
            pattern = ['Conv', 'Conv', 'Conv'],
            edges = [[0, 1], [1, 2], [0, 2]]

        描述了一个类似这样的树形结构:

            Conv -+- Conv -+- Conv
                  |        |
                  ----------

        第二个例子:
        pt = PatternTree(
                patterns = [lambda x: x.is_computing_op, 'Softplus', 'Tanh', 'Mul']
                edges = [[0, 1], [1, 2], [2, 3], [0, 3]])

            pt create an abstract tree pattern of:
                                            --- 'Softplus'   ---   'Tanh' --
            lambda x: x.is_computing_op --- +                              + --- 'Mul'
                                            ---     ---     ---    ---    --

        错误的例子:
            pattern = ['Conv', 'Conv', 'Conv'],
            edges = [[0, 1], [1, 2], [2, 0]]
        因为图中存在循环结构而无法检索
        """
        for idx, node_pattern in enumerate(node_patterns):
            if isinstance(node_pattern, str):
                node_patterns[idx] = TypeExpr(node_pattern)
            elif not isinstance(node_pattern, Callable):
                raise TypeError(f'Can not create Pattern with node pattern {str(node_pattern)} it is not callable.')

        for edge in edges:
            if not isinstance(edge, tuple) and not isinstance(edge, list):
                raise TypeError(f'Can not create Pattern with edge {str(edge)} it is not tuple or list.')
            if len(edge) != 2:
                raise ValueError(
                    f'Can not create Pattern with edge {str(edge)} '
                    f'it should contains exact 2 elements, however {len(edge)} was given.'
                )
            sp, ep = edge
            if not isinstance(sp, int) or not isinstance(ep, int):
                raise TypeError(f'Can not create Pattern was given edge {[str(sp), str(ep)]}, expect int value here.')

        self.order, self.output_table, self.input_table = self.compile(node_patterns=node_patterns, edges=edges)
        self.argsort_order = sorted([(_, idx) for idx, _ in enumerate(self.order)])
        self.argsort_order = [idx for _, idx in self.argsort_order]
        self.node_patterns = node_patterns

    def compile(self, node_patterns: List[Callable], edges: List[List[int]]):
        """Pattern Compile.
        1. Do topological Sort on given pattern.
        2. Reorganize pattern by topological order.
        """

        # prepare for topological sort
        visited = [False for _ in node_patterns]
        num_of_inputs = [0 for _ in node_patterns]
        forward_table = [set() for _ in node_patterns]
        backward_table = [set() for _ in node_patterns]
        roots = []

        for edge in edges:
            sp, ep = edge
            if sp >= len(node_patterns) or sp < 0:
                raise IndexError(
                    f'Can not Compile Pattern, Edge {edge} Out of Node Range, '
                    f'Except Value between 0 and {len(node_patterns) - 1}, however {sp} was given.'
                )
            if ep >= len(node_patterns) or ep < 0:
                raise IndexError(
                    f'Can not Compile Pattern, Edge {edge} Out of Node Range, '
                    f'Except Value between 0 and {len(node_patterns) - 1}, however {ep} was given.'
                )
            forward_table[sp].add(ep)
            backward_table[ep].add(sp)
            num_of_inputs[ep] += 1

        # initialization
        pop_list, sort_ret = deque(), []
        for idx, n_input in enumerate(num_of_inputs):
            if n_input == 0:
                pop_list.append(idx)
                roots.append(idx)

        # topological sort
        for _ in range(len(visited)):
            if len(pop_list) == 0:
                break
            current = pop_list.popleft()

            for next in forward_table[current]:
                num_of_inputs[next] -= 1
                if num_of_inputs[next] == 0:
                    pop_list.append(next)

            visited[current] = True
            sort_ret.append(current)

        if all(visited):
            if len(roots) > 1:
                ppq_warning('More than 1 pattern root was found, Complext Pattern might cause memory overflow ...')
            return sort_ret, forward_table, backward_table
        else:
            raise RuntimeError('Topological Sort failed. Some node can not be sorted (might due to circular reference)')


class TypeExpr(Callable):
    def __init__(self, type: str) -> None:
        self.type = type
        super().__init__()

    def __call__(self, op: Operation) -> bool:
        return op.type == self.type


class PatternMatchHelper:
    @staticmethod
    def match_burte_force(
        graph: BaseGraph, pattern: GraphPattern, exclusive: bool, max_candidates: int = 1000000
    ) -> List[List[Operation]]:
        """暴力子图模式匹配 这是 PPQ 0.6.6 更新的内容
        在 0.6.6 之前，我们使用具有不确定性的贪心匹配算法，但是考虑到实际应用中的问题
        在 0.6.6 版本之后，我们将其修改为枚举匹配。

        子图匹配问题是一个 NP-Hard 的问题，不存在多项式时间复杂度的解法。
        你需要给出一个模式子图，match_burte_force 方法将在 graph 中对模式子图进行匹配。

        PPQ 使用了非递归的算法完成上述匹配，其最坏时间和空间复杂度大概都是 O(NM^k)
        其中 N 是母图节点个数，M 是子图节点个数，k 是母图的最大出度

        对于存在二义性子图模式，匹配复杂度将指数级增长；为了限制算法执行时间，当匹配到多于
        max_candidates 个模式子图时，算法强制停机，并报错返回。

        实际使用中的时间复杂度更加接近于 O(NM)

        参数 exclusive 指定了是否需要进行精确匹配。在精确匹配模式下：
            1. 不允许模式子图中除根节点外的其他节点有来自模式子图以外节点的输入
            2. 不允许模式子图中除叶节点外的其他节点有朝向模式子图以外节点的输出

        Example:
            pt = PatternTree(
                patterns = [lambda x: x.is_computing_op, 'Softplus', 'Tanh', 'Mul']
                edges = [[0, 1], [1, 2], [2, 3], [0, 3]])

            pt create an abstract tree pattern of:
                                            --- 'Softplus'   ---   'Tanh' --
            lambda x: x.is_computing_op --- +                              + --- 'Mul'
                                            ---     ---     ---    ---    --

        """

        def is_linked(upstream_op: Operation, downstream_op: Operation) -> bool:
            if upstream_op is None or downstream_op is None:
                return True
            return downstream_op in graph.get_downstream_operations(upstream_op)

        node_order = pattern.order
        matched_patterns = []

        # match root from graph, further pattern matching will start from root.
        for operation in graph.operations.values():
            root_idx = node_order[0]
            if pattern.node_patterns[root_idx](operation):
                matched_patterns.append([operation] + [None for _ in range(len(node_order) - 1)])

        for idx in node_order[1:]:
            node_candidates, next_generation = [], []
            for operation in graph.operations.values():
                if pattern.node_patterns[idx](operation):
                    node_candidates.append(operation)

            for matched_pattern in matched_patterns:
                for operation in node_candidates:
                    is_pattern_root = len(pattern.input_table[idx]) == 0
                    link_check, exclusive_check, duplicated_check = True, True, True

                    for upstream_idx in pattern.input_table[idx]:
                        upstream_op = matched_pattern[upstream_idx]

                        if not is_linked(upstream_op=upstream_op, downstream_op=operation):
                            link_check = False

                    if (operation in matched_pattern) and link_check:
                        duplicated_check = False

                    if (not is_pattern_root and exclusive) and link_check:
                        matched_set = set(matched_pattern)
                        for op in graph.get_upstream_operations(operation):
                            if op not in matched_set:
                                exclusive_check = False
                        if len(pattern.input_table[idx]) != len(set(graph.get_upstream_operations(operation))):
                            exclusive_check = False

                    if link_check and duplicated_check and exclusive_check:
                        generated = matched_pattern.copy()
                        generated[idx] = operation
                        next_generation.append(generated)

            matched_patterns = next_generation
            if len(matched_patterns) > max_candidates:
                raise OverflowError('Too many candidate patterns. Simplify your pattern first.')

            if len(matched_patterns) == 0:
                break

        # final exclusive check
        if exclusive:
            filtered = []
            for matched_pattern in matched_patterns:
                check_flag = True
                for idx, operation in enumerate(matched_pattern):
                    if len(pattern.output_table[idx]) != 0:
                        if not all([op in matched_pattern for op in graph.get_downstream_operations(operation)]):
                            check_flag = False
                if check_flag:
                    filtered.append(matched_pattern)
            matched_patterns = filtered
        return matched_patterns


class SearchableGraph(GraphCommandProcessor):
    """PPQ Graph Searching Engine.

    Args:
        GraphCommandProcessor ([type]): [description]
    """

    def __init__(self, graph: Union[BaseGraph, Callable]) -> None:
        super().__init__(graph)
        self._cache = {}

    def process(self, command: GraphCommand) -> Any:
        if isinstance(command, TraversalCommand):
            if command.command_type == GraphCommandType.TRAVERSAL_PATTERN_MATCHING:
                return self.path_matching(
                    sp_expr=command._sp_expr,
                    rp_expr=command._rp_expr,
                    ep_expr=command._ep_expr,
                    direction=command._direction,
                )
            else:
                return self.opset_matching(
                    sp_expr=command._sp_expr,
                    rp_expr=command._rp_expr,
                    ep_expr=command._ep_expr,
                    direction=command._direction,
                )
        else:
            raise TypeError(
                'To execute a traversal-based pattern matching, a TraversalCommand is required here.\n'
                'Initialize your own TraversalCommand instance for invoking '
                'this function rather than using plain GraphCommand Please.'
            )

    def _acceptable_command_types(self) -> List[GraphCommandType]:
        return [
            GraphCommandType.TRAVERSAL_PATTERN_MATCHING,
            GraphCommandType.TRAVERSAL_OPSET_MATCHING,
            GraphCommandType.CONCAT_MATCHING,
            GraphCommandType.ACTIVATION_MATCHING,
        ]

    def _path_matching(
        self, start_point: Operation, rp_expr: RelayPattern, ep_expr: PointPattern, direction: str = 'up'
    ) -> List[Path]:
        # memoization based searching.
        if start_point in self._cache:
            return self._cache[start_point]

        # find next operations with given direction
        if direction == 'up':
            following_ops = self.graph.get_upstream_operations(start_point)
        else:
            following_ops = self.graph.get_downstream_operations(start_point)

        ret_collection = []
        for op in following_ops:
            # if operation is a valid end point, add it to path and stop further searching.
            if ep_expr(op):
                ret_collection.append(Path(start_point).append(op))

            else:
                # if operation is not a valid relay point, end searching here.
                if not rp_expr(start_point, op):
                    continue

                # searching following operations.
                for path in self._path_matching(start_point=op, rp_expr=rp_expr, ep_expr=ep_expr, direction=direction):
                    ret_collection.append(path.copy().append_left(start_point))

        self._cache[start_point] = ret_collection
        return ret_collection

    def _opset_matching(
        self, start_point: Operation, rp_expr: RelayPattern, ep_expr: PointPattern, direction: str = 'up'
    ) -> OperationSet:
        # memoization based searching.
        if start_point in self._cache:
            return self._cache[start_point]

        # begin searching.
        ret_collection = OperationSet()

        # find next operations with given direction
        if direction == 'up':
            following_ops = self.graph.get_upstream_operations(start_point)
        else:
            following_ops = self.graph.get_downstream_operations(start_point)

        # new feature with ppq 0.6.5, if ep_expr is None, means search until mismatch.
        if len(following_ops) == 0 and ep_expr is None:
            return ret_collection.add(start_point)

        for op in following_ops:
            # if operation is a valid end point, add it to path and stop further searching.
            if ep_expr is not None and ep_expr(op):
                ret_collection.update([start_point, op])

            else:
                # if operation is not a valid relay point, end searching here.
                if not rp_expr(start_point, op):
                    # new feature with ppq 0.6.5, if ep_expr is None, means search until mismatch.
                    if ep_expr is None:
                        ret_collection.add(start_point)
                    continue

                # searching following operations.
                further_result = self._opset_matching(
                    start_point=op, rp_expr=rp_expr, ep_expr=ep_expr, direction=direction
                )

                if len(further_result) > 0:
                    ret_collection.update(further_result)
                    ret_collection.add(start_point)

        self._cache[start_point] = ret_collection
        return ret_collection

    def path_matching(self, sp_expr: Callable, rp_expr: Callable, ep_expr: Callable, direction: str) -> List[Path]:
        """path_matching is used for path searching on the graph, and complete
        paths will be returned. Note that it's possible for results to overflow
        when there are numerous matchings.

        path_matching 用于图检索，匹配完整路径
            一个最简单的例子：
                sp_expr = lamdba x: x.type == 'Conv'
                rp_expr = lamdba x, y: y.type == 'Relu'
                ep_expr = lamdba x: x.type == 'Conv'
                limitation = None
                direction = 'down'
            该指令检索出从任意 Conv 出发，到任意 Conv 的所有可行路径
            其中路径上含有任意多个 Relu 节点，并且只能包含 Relu

            你需要注意到 rp_expr 是一个二元表达式，其中 x 代表起点，y 代表终点
        Args:
            sp_expr (Union[Pattern, Callable]):
                start point expression, 用于匹配检索起点的表达式
            rp_expr (Union[Pattern, Callable]):
                relay point expression, 用于匹配检索中继点的表达式
            ep_expr (Union[Pattern, Callable]):
                end point expression, 用于匹配检索终点的表达式
            limitation (Union[Limitation, Callable]):
                用于过滤路径的表达式
            direction (str, optional):
                图检索方向，search direction: up, down. Defaults to 'down'.
            greedy (bool, optional):
                是否执行贪心匹配，设置为True则尝试匹配直到最后一个end point.
                whether to search greedily
        """
        if direction not in {'up', 'down'}:
            raise KeyError("traversal_direction must be one of {'up', 'down'}")

        _ret_collection = []
        for operation in self.graph.operations.values():
            # filter out operation which can not be a start point.
            if not sp_expr(operation):
                continue

            # match patterns, add patterns towards result collection.
            matchings = self._path_matching(
                start_point=operation, rp_expr=rp_expr, ep_expr=ep_expr, direction=direction
            )
            _ret_collection.extend(matchings)

        # clear cache
        self._cache.clear()

        # use limitation to filter out invalid path.
        return _ret_collection

    def opset_matching(
        self, sp_expr: Callable, rp_expr: Callable, ep_expr: Callable, direction: str = 'up'
    ) -> OperationSet:
        """opset_matching is used for operator set searching, it returns
        relevant op set instead of specific paths, should be used when results
        of path_matching overflow.

        opset_matching 用于图检索，匹配算子集合（无序）
            一个最简单的例子：
                sp_expr = lamdba x: x.type == 'Conv'
                rp_expr = lamdba x, y: y.type == 'Relu'
                ep_expr = lamdba x: x.type == 'Conv'
                limitation = None
                direction = 'down'
            该指令检索出从任意 Conv 出发，经过任意多个Relu，到任意 Conv 的所有相关算子
            注意返回结果是无序的，相比于 path matching，opset matching 的性能更高。

            你需要注意到 rp_expr 是一个二元表达式，其中 x 代表起点，y 代表终点

            在极端情况下，path matching 的结果是无法返回的（由于其结果数可能过多）
            此时应当使用 opset matching
        Args:
            sp_expr (Union[Pattern, Callable]):
                start point expression, 用于匹配检索起点的表达式
            rp_expr (Union[Pattern, Callable]):
                relay point expression, 用于匹配检索中继点的表达式
            ep_expr (Union[Pattern, Callable]):
                end point expression, 用于匹配检索终点的表达式
            limitation (Union[Limitation, Callable]):
                用于过滤路径的表达式
            direction (str, optional):
                图检索方向，up, down. Defaults to 'down'.
            greedy (bool, optional):
                是否执行贪心匹配，设置为True则尝试匹配直到最后一个end point.
        """
        ret_collection, candidates = OperationSet(), OperationSet()
        for operation in self.graph.operations.values():
            # filter out operation which can not be a start point.
            if sp_expr(operation):
                candidates.add(operation)

        for operation in candidates:
            # match patterns, add patterns towards result collection.
            partial_matchings = self._opset_matching(
                start_point=operation, rp_expr=rp_expr, ep_expr=ep_expr, direction=direction
            )

            if len(partial_matchings) > 0:
                ret_collection.update(partial_matchings)

        # clear cache
        self._cache.clear()
        return ret_collection

    def activation_matching(
        self,
        start_op_types: Set[str],
        end_types: Set[str],
    ) -> Dict[str, List[str]]:
        matchings = self.path_matching(
            sp_expr=lambda op: op.type in start_op_types,
            rp_expr=lambda x, y: False,  # must have no relay operation.
            ep_expr=lambda op: op.type in end_types,
            direction='down',
        )

        activation_matchings = defaultdict(list)
        for path in matchings:
            op, act = path[0], path[-1]
            activation_matchings[op.name].append(act.name)
        return dict(activation_matchings)

    def concat_matching(self, relay_pattern: Callable, end_pattern: Callable) -> Dict[str, List[str]]:
        matchings = self.path_matching(
            sp_expr=lambda op: op.type in {'Concat'}, rp_expr=relay_pattern, ep_expr=end_pattern, direction='up'
        )

        concat_matchings = defaultdict(list)
        for path in matchings:
            concat, op = path[0], path[-1]
            concat_matchings[concat.name].append(op.name)
        return dict(concat_matchings)

    def pattern_matching(
        self, patterns: List[Callable], edges: List[List[int]], exclusive: bool = True
    ) -> List[List[Operation]]:
        """暴力子图模式匹配 这是 PPQ 0.6.6 更新的内容
        在 0.6.6 之前，我们使用具有不确定性的贪心匹配算法，但是考虑到实际应用中的问题
        在 0.6.6 版本之后，我们将其修改为枚举匹配。

        子图匹配问题是一个 NP-Hard 的问题，不存在多项式时间复杂度的解法。
        你需要给出一个模式子图，match_burte_force 方法将在 graph 中对模式子图进行匹配。

        PPQ 使用了非递归的算法完成上述匹配，其最坏时间和空间复杂度大概都是 O(NM^k)
        其中 N 是母图节点个数，M 是子图节点个数，k 是母图的最大出度

        对于存在二义性子图模式，匹配复杂度将指数级增长；为了限制算法执行时间，当匹配到多于
        max_candidates 个模式子图时，算法强制停机，并报错返回。

        实际使用中的时间复杂度更加接近于 O(NM)

        参数 exclusive 指定了是否需要进行精确匹配。在精确匹配模式下：
            1. 不允许模式子图中除根节点外的其他节点有来自模式子图以外节点的输入
            2. 不允许模式子图中除叶节点外的其他节点有朝向模式子图以外节点的输出

        Example:
            pt = PatternTree(
                patterns = [lambda x: x.is_computing_op, 'Softplus', 'Tanh', 'Mul']
                edges = [[0, 1], [1, 2], [2, 3], [0, 3]])

            pt create an abstract tree pattern of:
                                            --- 'Softplus'   ---   'Tanh' --
            lambda x: x.is_computing_op --- +                              + --- 'Mul'
                                            ---     ---     ---    ---    --

        """
        return PatternMatchHelper().match_burte_force(
            graph=self.graph, pattern=GraphPattern(node_patterns=patterns, edges=edges), exclusive=exclusive
        )



from .base.command import GraphCommand, GraphCommandType
from .base.graph import BaseGraph, GraphBuilder, GraphExporter, Operation, OperationExporter, Opset, Variable
from .base.opdef import OperationBase, OpSocket, VLink
from .deploy import RunnableGraph
from .morph import GraphFormatter, GraphMerger, GraphReplacer
from .processer import DefaultGraphProcessor, GraphCommandProcessor
from .quantize import DeviceSwitchOP, QuantableGraph, QuantableOperation, QuantableVariable
from .search import GraphPattern, Path, PatternMatchHelper, SearchableGraph, TraversalCommand
from .training import TrainableGraph



"""
## PPQ Foundation Library - PFL

PPQ 基础类库

PFL is a collection of basic classes and functions that provides fundamental functionalities.

    - Parser: Get a network parser.

    - Exporter: According to given platform, get a network exporter.

    - OperationForwardFunction: According to given platform and optype, get a forward function.

    - Dispatcher: Get a network dispatcher.

    - FloatingQuantizationConfig: Get a TensorQuantizationConfig for FP8 Quantization.

    - LinearQuantizationConfig: Get a TensorQuantizationConfig for INT8 Quantization.

    - QuantStub: Get a QuantStub class instance.

    - Quantizer: Get a Quantizer corresponding to given platform.

    - Observer: Get a Tensor Observer, which is bound to given TensorQuantizationConfig.

    - Pipeline: Build Optimization Pipeline.

    - QuantFunction: Get PPQ Default Quantize Function.

PFL also provides a set of functions to register Quantizer, Parser, Exporter to esp_ppq.

    - register_network_quantizer

    - register_network_parser

    - register_network_exporter

    - register_operation_handler

    - register_calibration_observer

"""

from .extension import (
    register_calibration_observer,
    register_network_exporter,
    register_network_parser,
    register_network_quantizer,
    register_operation_handler,
)
from .quant import (
    Dispatcher,
    Exporter,
    FloatingQuantizationConfig,
    LinearQuantizationConfig,
    Observer,
    OperationForwardFunction,
    ParameterQuant,
    Parser,
    Pipeline,
    QuantFunction,
    Quantizer,
    TensorQuant,
)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: Dl

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class StringStringEntry(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = StringStringEntry()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsStringStringEntry(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # StringStringEntry
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # StringStringEntry
    def Key(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # StringStringEntry
    def Value(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

def StringStringEntryStart(builder):
    builder.StartObject(2)

def Start(builder):
    StringStringEntryStart(builder)

def StringStringEntryAddKey(builder, key):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(key), 0)

def AddKey(builder, key):
    StringStringEntryAddKey(builder, key)

def StringStringEntryAddValue(builder, value):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(value), 0)

def AddValue(builder, value):
    StringStringEntryAddValue(builder, value)

def StringStringEntryEnd(builder):
    return builder.EndObject()

def End(builder):
    return StringStringEntryEnd(builder)


class StringStringEntryT(object):

    # StringStringEntryT
    def __init__(self):
        self.key = None  # type: str
        self.value = None  # type: str

    @classmethod
    def InitFromBuf(cls, buf, pos):
        stringStringEntry = StringStringEntry()
        stringStringEntry.Init(buf, pos)
        return cls.InitFromObj(stringStringEntry)

    @classmethod
    def InitFromPackedBuf(cls, buf, pos=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, pos)
        return cls.InitFromBuf(buf, pos+n)

    @classmethod
    def InitFromObj(cls, stringStringEntry):
        x = StringStringEntryT()
        x._UnPack(stringStringEntry)
        return x

    # StringStringEntryT
    def _UnPack(self, stringStringEntry):
        if stringStringEntry is None:
            return
        self.key = stringStringEntry.Key()
        self.value = stringStringEntry.Value()

    # StringStringEntryT
    def Pack(self, builder):
        if self.key is not None:
            key = builder.CreateString(self.key)
        if self.value is not None:
            value = builder.CreateString(self.value)
        StringStringEntryStart(builder)
        if self.key is not None:
            StringStringEntryAddKey(builder, key)
        if self.value is not None:
            StringStringEntryAddValue(builder, value)
        stringStringEntry = StringStringEntryEnd(builder)
        return stringStringEntry



import argparse
import json
import os

import tensorrt as trt

TRT_LOGGER = trt.Logger()

EXPLICIT_BATCH = 1 << (int)(trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH)


def GiB(val):
    return val * 1 << 30


def json_load(filename):
    with open(filename) as json_file:
        data = json.load(json_file)
    return data


def setDynamicRange(network, json_file):
    """Sets ranges for network layers."""
    quant_param_json = json_load(json_file)
    act_quant = quant_param_json["act_quant_info"]

    for i in range(network.num_inputs):
        input_tensor = network.get_input(i)
        if act_quant.__contains__(input_tensor.name):
            value = act_quant[input_tensor.name]
            tensor_max = abs(value)
            tensor_min = -abs(value)
            input_tensor.dynamic_range = (tensor_min, tensor_max)

    for i in range(network.num_layers):
        layer = network.get_layer(i)

        for output_index in range(layer.num_outputs):
            tensor = layer.get_output(output_index)

            if act_quant.__contains__(tensor.name):
                value = act_quant[tensor.name]
                tensor_max = abs(value)
                tensor_min = -abs(value)
                tensor.dynamic_range = (tensor_min, tensor_max)
            else:
                print("\033[1;32m%s\033[0m" % tensor.name)


def build_engine(onnx_file, json_file, engine_file):
    builder = trt.Builder(TRT_LOGGER)
    network = builder.create_network(EXPLICIT_BATCH)

    config = builder.create_builder_config()

    # If it is a dynamic onnx model , you need to add the following.
    # profile = builder.create_optimization_profile()
    # profile.set_shape("input_name", (batch, channels, min_h, min_w), (batch, channels, opt_h, opt_w), (batch, channels, max_h, max_w))
    # config.add_optimization_profile(profile)

    parser = trt.OnnxParser(network, TRT_LOGGER)
    config.max_workspace_size = GiB(1)

    if not os.path.exists(onnx_file):
        quit('ONNX file {} not found'.format(onnx_file))

    with open(onnx_file, 'rb') as model:
        if not parser.parse(model.read()):
            print('ERROR: Failed to parse the ONNX file.')
            for error in range(parser.num_errors):
                print(parser.get_error(error))
            return None

    config.set_flag(trt.BuilderFlag.INT8)

    setDynamicRange(network, json_file)

    engine = builder.build_engine(network, config)

    with open(engine_file, "wb") as f:
        f.write(engine.serialize())


if __name__ == '__main__':
    # Add plugins if needed
    # import ctypes
    # ctypes.CDLL("libmmdeploy_tensorrt_ops.so")
    parser = argparse.ArgumentParser(description='Writing qparams to onnx to convert tensorrt engine.')
    parser.add_argument('--onnx', type=str, default=None)
    parser.add_argument('--qparam_json', type=str, default=None)
    parser.add_argument('--engine', type=str, default=None)
    arg = parser.parse_args()

    build_engine(arg.onnx, arg.qparam_json, arg.engine)
    print("\033[1;32mgenerate %s\033[0m" % arg.engine)



import json
import os
import sys
from typing import Any, Dict, List, Sequence, Tuple, Union

import numpy as np
import torch

from esp_ppq.core import (
    ONNX_VERSION,
    PPQ_CONFIG,
    DataType,
    QuantizationStates,
    convert_any_to_numpy,
    convert_any_to_torch_tensor,
)
from esp_ppq.IR import BaseGraph, GraphExporter, Operation, OperationExporter, Variable
from esp_ppq.IR.quantize import QuantableOperation
from esp_ppq.log import NaiveLogger
from esp_ppq.quantization.qfunction.linear import PPQLinearQuant_toInt

from .espdl import helper
from .espdl.espdl_streaming import AutoStreamingPattern, StreamingTable, insert_streaming_nodes
from .espdl.espdl_typedef import REDUCE_OP_SET, ExporterPatternInfo, LayoutAnnotation
from .espdl.export_patterns import (
    AddLUTPattern,
    FuseReluLikePattern,
    InsertDequantNodePattern,
    InsertPreNodeOfMatMulPattern,
    InsertQuantNodePattern,
    InsertQuantTypePattern,
    InsertRequantNodePattern,
    QuantRNNPattern,
    QuantVariableToIntPattern,
    ResetParamLayoutPattern,
)
from .espdl.layout_patterns import (
    reset_graph_layout,
    transpose_shape,
)
from .onnx_exporter import OP_CONVERTERS

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), ".")))

logger = NaiveLogger.get_logger("ESPDL")


def convert_value(value: Union[int, float, np.ndarray, torch.Tensor]) -> Any:
    if type(value) in {int, float}:
        return value
    else:
        value = convert_any_to_numpy(value, accept_none=True)
        if value is None:
            return value  # SOI config has Nona as its scale and
        return value.tolist()


def set_streaming_input_shape(origin_shape, shape, perm) -> None:
    """Set streaming input shape for graph inputs.

    Args:
        graph (BaseGraph): Processing graph.
        streaming_input_shape (Dict[str, List[Any]]): A dictionary mapping input variable names to their streaming shapes.
    """
    diff_dim = 0
    for i in range(len(shape)):
        if shape[i] != origin_shape[i]:
            StreamingTable().set_input_frame_axis(perm[i])
            StreamingTable().set_input_frame_num(shape[i])
            diff_dim += 1
    if diff_dim != 1:
        raise ValueError(f"Only one dimension can be different for streaming input shape. {origin_shape} -> {shape}")


class EspdlExporter(GraphExporter):
    """
    The EspdlExporter is used to export computational graphs into the esp-dl standard format.
    The export logic of any exporter is performed in-place, meaning it will modify the incoming computational graph object in-place.
    Therefore, you need to manually clone the computational graph before exporting.
    """

    def __init__(self) -> None:
        super().__init__()

    def export(
        self,
        file_path: str,
        graph: BaseGraph,
        config_path: str = None,
        model_version: int = 0,
        values_for_test: Dict[str, Dict[str, torch.Tensor]] = None,
        export_config: bool = True,
        encrypt_data: bool = False,
        int16_lut_step: int = 256,
        metadata_props: Dict[str, str] = None,
        streaming_table: Dict[str, Dict[str, Any]] = None,
        auto_streaming: bool = False,
        streaming_input_shape: List[Any] = None,
        **kwargs: Any,
    ):
        """Export model to flatbuffers file

        Args:
            file_path: flatbuffers file
            graph: ppq graph
            config_path: quantization config
            model_version (int): model version
            values_for_test (Dict[str, Dict[str, np.ndarray]]): the test values used to compare accuracy.
                                                            The input format is as follows:
                {
                    'inputs': {
                        'input_0_name': np.ndarray
                        ......
                        'input_n_name': np.ndarray
                    },
                    'outputs': {
                        'output_0_name': np.ndarray
                        ......
                        'output_n_name': np.ndarray
                    },
                }
            int16_lut_step (int): -1: do not add lut for int16, otherwise add lut for int16 with given step
            metadata_props (Dict[str, str]): metadata properties
        """

        # during export we will remove all boundary operations from graph.
        # we do not want to change the structure of original graph,
        # so there have to take a clone of it.
        graph = graph.copy()

        # In prepare stage, run all graph pattern
        # fuse Conv and Relu and insert quant node if necessary
        exporter_patterns = {
            "pre_patterns": [
                FuseReluLikePattern,
                InsertQuantNodePattern,
                InsertRequantNodePattern,
                InsertDequantNodePattern,
                InsertPreNodeOfMatMulPattern,
            ],
            "post_patterns": [
                InsertQuantTypePattern,
                QuantVariableToIntPattern,
                ResetParamLayoutPattern,
                QuantRNNPattern,
            ],
        }

        graph = self.prepare_graph(
            graph, exporter_patterns, int16_lut_step, streaming_table, auto_streaming, streaming_input_shape
        )
        file_base_name, _ = os.path.splitext(file_path)

        # if a valid config path is given, export quantization config to there.
        if export_config:
            if not config_path:
                config_path = file_base_name + ".json"
            self.export_quantization_config(config_path, graph)

        model = self.export_graph(
            graph=graph,
            model_version=model_version,
            values_for_test=values_for_test,
            metadata_props=metadata_props,
            **kwargs,
        )

        # Export the information of quantized espdl model.
        if export_config:
            info_file_path = file_base_name + ".info"
            with open(info_file_path, "w") as file_info:
                file_info.write(
                    helper.printable_graph(
                        model,
                        print_initializer_value=True,
                        print_value_info=True,
                        print_test_value=True,
                    )
                )

        key = helper.save(
            model,
            file_path,
            encrypt_data,  # if True, encrypt data for security
        )
        if key:
            logger.info(f"AES 128-bit secret key: {key}")

        ExporterPatternInfo().reset()

    def prepare_graph(
        self,
        graph: BaseGraph,
        exporter_patterns: Dict[str, List[OperationExporter]],
        int16_lut_step: int,
        streaming_table: Dict[str, Dict[str, Any]] = None,
        auto_streaming: bool = False,
        streaming_input_shape: List[Any] = None,
    ) -> BaseGraph:
        """Prepare your graph for exporting.

        There are many works to do with your graph:

            1. Insert Quant and Dequant operation within your graph.

            2. Remove all unnecessary activations.

            3. Quantize all parameters of your graph, convert them to int8.

        Args:
            graph (BaseGraph): Processing Graph

        Returns:
            BaseGraph: Processed Graph
        """
        self.convert_operation_from_opset11_to_opset18(graph)

        # before we can export them, we firstly convert all ops to proper format. Copy from onnx exporter
        for op in [_ for _ in graph.topological_sort()]:
            if op.type in OP_CONVERTERS:
                exporter = OP_CONVERTERS[op.type]()
                assert isinstance(exporter, OperationExporter), (
                    f"Expected an OpExporter here, however {type(exporter)} was given."
                )
                op = exporter.export(op=op, graph=graph)

        for pattern in exporter_patterns.get("pre_patterns", {}):
            exporter = pattern()
            for op in graph.topological_sort():
                exporter.export(op=op, graph=graph)

        # reset Conv layout from NCHW to NHWC and insert transpose node if necessary
        reset_graph_layout(graph)

        for pattern in exporter_patterns.get("post_patterns", {}):
            exporter = pattern()
            for op in graph.topological_sort():
                exporter.export(op=op, graph=graph)

        info = ExporterPatternInfo()
        if auto_streaming or streaming_table is not None:
            logger.info("Inserting streaming nodes into graph...")
            StreamingTable().append(streaming_table)
            if streaming_input_shape != None:
                logger.info("Setting streaming input shape...")
                if isinstance(streaming_input_shape, dict):
                    for var_name in streaming_input_shape:
                        shape = streaming_input_shape[var_name]
                        if var_name in graph.inputs:
                            set_streaming_input_shape(
                                graph.inputs[var_name].shape, shape, info.get_var_permute(var_name)
                            )
                            graph.inputs[var_name].shape = shape
                        else:
                            logger.warning(f"Input variable {var_name} not found in graph inputs.")
                elif isinstance(streaming_input_shape, list) and len(graph.inputs) == 1:
                    for var in graph.inputs.values():
                        set_streaming_input_shape(var.shape, streaming_input_shape, info.get_var_permute(var.name))
                        var.shape = streaming_input_shape
                    # graph.inputs[graph.inputs.keys[0]].shape = streaming_input_shape
                else:
                    raise ValueError("Invalid streaming_input_shape provided.")
            else:
                raise ValueError("streaming_input_shape must be provided when auto_streaming is enabled.")

            # insert streaming node pattern
            if auto_streaming:
                exporter = AutoStreamingPattern()
                for op in graph.topological_sort():
                    exporter.export(op=op, graph=graph)
            insert_streaming_nodes(graph)
            StreamingTable().reset()

        # prepare LUT
        exporter = AddLUTPattern(int16_step=int16_lut_step)
        for op in graph.topological_sort():
            exporter.export(op=op, graph=graph)

        for variable in graph.variables.values():
            if variable.is_parameter or len(variable.name) == 0:
                continue

            perm = info.get_var_permute(variable.name)
            exponents = info.get_var_exponents(variable.name)
            layout = info.get_var_layout(variable.name)
            if exponents and perm:
                logger.debug(f"{variable.name} perm: {perm}, exponents: {exponents}, layout:{layout}")
            elif not perm:
                logger.warning(f"{variable.name} does not bind perm parameter")
            elif not exponents and variable.dtype in {DataType.INT8, DataType.INT16}:
                logger.warning(f"{variable.name} does not bind exponents parameter")

        for op in graph.topological_sort():
            if op.type in ["Conv", "Gemm", "MatMul", "Mul"]:
                input0_exponent = info.get_var_exponents(op.inputs[0].name)
                input1_exponent = info.get_var_exponents(op.inputs[1].name)
                output_exponent = info.get_var_exponents(op.outputs[0].name)
                # By default, it is per-tensor. Currently, esp-dl does not support per-channel.
                if input0_exponent and input1_exponent and output_exponent:
                    if (output_exponent[0] - input0_exponent[0] - input1_exponent[0]) < 0:
                        logger.error(
                            f"When deploying with esp-dl, the calculation result of the {op.name}(type: {op.type}) "
                            "operator will cause an exception. Please adjust the model to ensure "
                            "that (output_exponent - input0_exponent - input1_exponent) >= 0."
                        )

        return graph

    def export_graph(
        self,
        graph: BaseGraph,
        model_version: int = 0,
        values_for_test: Dict[str, Dict[str, torch.Tensor]] = None,
        metadata_props: Dict[str, str] = None,
        **kwargs: Any,
    ) -> bytes:
        """
        Convert a PPQ IR to Onnx IR.
        This export will only convert PPQ Op and var to onnx, all quantization configs will be skipped.

        This function will try to keep the opset version of your graph unchanged.
        However if the opset is not given, ppq will convert it to with the global parameter esp_ppq.core.ONNX_EXPORT_OPSET.
        """
        name = graph._name
        if not name:
            name = f"{PPQ_CONFIG.NAME} - v({PPQ_CONFIG.VERSION})"

        # Ready to export espdl graph defination.
        _inputs, _outputs, _initilizers, _nodes, _value_info = [], [], [], [], []
        pattern_info = ExporterPatternInfo()

        # Add LUT to graph
        _initilizers += self.build_lut_proto(graph, pattern_info)

        for op in graph.topological_sort():
            _nodes.append(self.build_operator_proto(op))

        for variable in graph.variables.values():
            tensor_proto = self.build_variable_proto(
                variable,
                pattern_info.get_var_exponents(variable.name, [0]),
                pattern_info.get_var_layout(variable.name, LayoutAnnotation.UNKNOWN),
                pattern_info.get_var_permute(variable.name),
            )
            if variable.name in graph.inputs:
                _inputs.append(tensor_proto)
            if variable.name in graph.outputs:
                _outputs.append(tensor_proto)
            if variable.is_parameter:
                _initilizers.append(tensor_proto)
            else:
                _value_info.append(tensor_proto)

        test_inputs_value, test_outputs_value = self.build_test_value_proto(values_for_test)

        graph_def = helper.make_graph(
            name=name,
            nodes=_nodes,
            inputs=_inputs,
            outputs=_outputs,
            initializer=_initilizers,
            value_info=_value_info,
            test_inputs_value=test_inputs_value,
            test_outputs_value=test_outputs_value,
        )

        metadata_props = helper.make_metadata_props(metadata_props)

        espdl_model = helper.make_model(
            graph=graph_def,
            producerName=PPQ_CONFIG.NAME,
            model_version=model_version,
            ir_version=graph._detail.get("ir_version", ONNX_VERSION),
            metadata_props=metadata_props,
        )

        return espdl_model

    def build_lut_proto(self, graph: BaseGraph, info: ExporterPatternInfo) -> List[int]:
        """
        Convert torch.Tensor to flatbuffer.Tensor and add them to graph.
        """
        _lut = []
        for op in graph.topological_sort():
            lut_name = op.attributes.get("lut", None)
            if lut_name:
                lut = info.get_lut(lut_name)
                onnx_dtype = DataType.convert_from_torch(lut.dtype).value
                lut_value = convert_any_to_numpy(lut).flatten()
                lut_proto = helper.make_tensor(
                    name=lut_name,
                    data_type=onnx_dtype,
                    dims=lut.shape,
                    vals=lut_value,
                    raw=True,
                    exponents=info.get_var_exponents(lut_name),
                )

                _lut.append(lut_proto)
        return _lut

    def build_test_value_proto(
        self, values_for_test: Dict[str, Dict[str, torch.Tensor]] = None
    ) -> Tuple[Sequence[int], Sequence[int]]:
        def quantize_and_transpose(var_name, tensor, pattern_info):
            if tensor.dim() == 0:
                tensor = tensor.unsqueeze(0)
            perm = pattern_info.get_var_permute(var_name)
            config = pattern_info.get_var_config(var_name)
            if perm:
                tensor = tensor.permute(perm)
            if config:
                tensor = PPQLinearQuant_toInt(tensor, config=config)

            return convert_any_to_numpy(tensor)

        pattern_info = ExporterPatternInfo()
        quant_values_for_test = {}
        if values_for_test is not None:
            if "inputs" not in quant_values_for_test:
                quant_values_for_test["inputs"] = {}
            if "outputs" not in quant_values_for_test:
                quant_values_for_test["outputs"] = {}

            for var_name in values_for_test.get("inputs", {}):
                tensor = quantize_and_transpose(var_name, values_for_test["inputs"][var_name], pattern_info)
                quant_values_for_test["inputs"][var_name] = tensor

            for var_name in values_for_test.get("outputs", {}):
                tensor = quantize_and_transpose(var_name, values_for_test["outputs"][var_name], pattern_info)
                quant_values_for_test["outputs"][var_name] = tensor

        return helper.make_graph_test_value(quant_values_for_test, pattern_info.var_exponents)

    def build_operator_proto(self, operation: Operation) -> int:
        """
        Convert PPQ Op to Onnx Operation
        An Op consumes zero or more Tensors, and produces zero or more Tensors.
        """
        attributes = operation.attributes
        for key in attributes:
            value = attributes[key]
            if isinstance(value, DataType):
                attributes[key] = value.value
            elif isinstance(value, np.dtype):
                value = DataType.convert_from_numpy(value)
                attributes[key] = value.value

            if isinstance(value, torch.Tensor):
                if value.numel() == 0:
                    attributes[key] = None
                elif value.numel() == 1:
                    attributes[key] = convert_any_to_numpy([value.item()])  # convert to 1d array
                else:
                    attributes[key] = convert_any_to_numpy(value)

        if PPQ_CONFIG.EXPORT_PPQ_INTERNAL_INFO:
            attributes["platform"] = operation.platform.name

        op_proto = helper.make_node(
            op_type=operation.type,
            inputs=[_.name for _ in operation.inputs],
            outputs=[_.name for _ in operation.outputs],
            name=operation.name,
            **attributes,
        )

        return op_proto

    def build_variable_proto(
        self,
        variable: Variable,
        exponent: List[int],
        layout: str,
        perm: Union[None, List[int]] = None,
    ) -> int:
        """
        Convert PPQ Variable to Onnx TensorProto, There are 2 different types of Tensor in Onnx:
            Variable: Represents a Tensor whose value is not known until inference-time.
            Constant: Represents a Tensor whose value is known.
        """
        # Parameter Varaible in PPQ, Constant Variable in Onnx
        if variable.is_parameter:
            if variable.value is not None:
                var_shape = variable.value.shape
                pytorch_dtype = variable.value.dtype
                onnx_dtype = DataType.convert_from_torch(pytorch_dtype).value

        # Non Parameter
        else:
            var_shape = variable.shape
            onnx_dtype = variable.dtype.value
            if var_shape:
                if len(var_shape) != len(perm):
                    logger.error(f"{variable.name} permute do not match shape")
                var_shape = transpose_shape(var_shape, perm)
            elif var_shape is not None and len(var_shape) == 0:
                var_shape = torch.tensor([1]).shape

        # if variable not in exponents, set exponent to 0
        var_exponents = exponent

        if not variable.is_parameter:
            tensor_proto = helper.make_tensor_value_info(
                name=variable.name,
                elem_type=onnx_dtype,
                shape=var_shape,
                exponents=var_exponents,
            )
        else:
            value = variable.value
            is_raw_format = False
            if isinstance(value, torch.Tensor):
                if value.numel() == 0:
                    value = []
                elif value.ndim >= 1:
                    value = convert_any_to_numpy(variable.value).flatten()
                    is_raw_format = True
                elif value.ndim == 0:  # Pytorch Scalar Type
                    value = [
                        value.item(),
                    ]  # it is fine for onnx, shape for this value will be []
                    var_shape = torch.tensor(value).shape
                    is_raw_format = True
            else:
                value = value  # value is python primary type.
            tensor_proto = helper.make_tensor(
                name=variable.name,
                data_type=onnx_dtype,
                dims=var_shape,
                vals=value,
                raw=is_raw_format,
                exponents=var_exponents,
                doc_string="layout ==> " + layout,
            )
        return tensor_proto

    def export_quantization_config(self, config_path: str, graph: BaseGraph):
        """Export Tensor Quantization Config to File(Json)."""

        render_buffer = {"configs": {}, "dispatchings": {}, "values": {}}

        # Render quantization config.
        for operation in graph.operations.values():
            if isinstance(operation, QuantableOperation):
                op_dict = {
                    var.name: {
                        "bit_width": config.num_of_bits,
                        "policy": config.policy.to_dict(),
                        "state": config.state.name,
                        "quant_min": config.quant_min,
                        "quant_max": config.quant_max,
                        "hash": config.__hash__(),
                        "dominator": config.dominated_by.__hash__(),
                    }
                    for config, var in operation.config_with_variable
                }

                for config, _ in operation.config_with_variable:
                    if config.dominated_by == config:
                        if config.state != QuantizationStates.FP32:
                            render_buffer["values"][config.__hash__()] = {
                                "scale": convert_value(config.scale),
                                "zero_point": convert_value(config.offset),
                            }

                render_buffer["configs"][operation.name] = op_dict
                render_buffer["dispatchings"][operation.name] = operation.platform.name

        with open(file=config_path, mode="w") as file:
            json.dump(render_buffer, file, indent=4)

    def convert_operation_from_opset11_to_opset18(self, graph: BaseGraph) -> None:
        """Convert your network from opset 11 standard towards opset 18 With
        Onnx definition, per-channel quant operation requires opset 13.

        Args:
            graph (BaseGraph): Processing graph.
        """
        # this func transform representation of certain op from opset 11 to 18
        for op in graph.operations.values():
            if op.type in ["Squeeze", "Unsqueeze"] + list(REDUCE_OP_SET):
                if "axes" not in op.attributes:
                    continue  # is already v13
                axes = convert_any_to_torch_tensor(op.attributes.pop("axes"), dtype=torch.int64)
                graph.create_variable(name=None, value=axes, is_parameter=True, dest_ops=[op])

            elif op.type == "Split":
                if "split" not in op.attributes:
                    continue  # split is already v13
                split = convert_any_to_torch_tensor(op.attributes.pop("split"), dtype=torch.int64)
                graph.create_variable(name=None, value=split, is_parameter=True, dest_ops=[op])



import copy
import json
import os
import sys
from typing import Any, Dict, List, Union

import numpy as np
import torch

from esp_ppq.core import (
    DataType,
    OperationQuantizationConfig,
    QuantizationProperty,
    QuantizationStates,
    QuantizationVisibility,
    SingletonMeta,
    TensorQuantizationConfig,
    convert_any_to_numpy,
    convert_any_to_torch_tensor,
)
from esp_ppq.IR import BaseGraph, Operation, OperationExporter, Variable
from esp_ppq.IR.quantize import QuantableOperation
from esp_ppq.log import NaiveLogger
from esp_ppq.parser.espdl.espdl_graph_utils import (
    fuse_downstream_operation,
    get_default_perm,
    get_inverse_transpose,
    insert_transpose_node,
    restore_origin_shape,
    transpose_shape,
)
from esp_ppq.parser.espdl.espdl_typedef import (
    ACTIVATION_OP_SET,
    ADD_LIKE_OP_SET,
    MATH_OP_SET,
    QUANT_OP_SET,
    ExporterPatternInfo,
)
from esp_ppq.utils.round import ppq_tensor_round

logger = NaiveLogger.get_logger("ESPDL")

CONV_STREAMING_OP_SET = {
    "Conv",
    "AveragePool",
    "MaxPool",
    "ConvTranspose",
}

BYPASS_OTHER_OP_SET = {
    "MatMul",
    "Gemm",
    "Flatten",
    "Reshape",
    "Squeeze",
    "Unsqueeze",
    "Concat",
    "Expand",
    "LayerNorm",
    "LayerNormalization",
    "Max",
    "Min",
    "TopK",
    "GRU",
    "LSTM",
    "InsertZeros",
}

STREADMING_BYPASS_OP_SET = ACTIVATION_OP_SET | MATH_OP_SET | QUANT_OP_SET | ADD_LIKE_OP_SET | BYPASS_OTHER_OP_SET


class StreamingTable(metaclass=SingletonMeta):
    _table: Dict[str, Dict[str, Any]] = {}
    _input_frame_axis: int = 1
    _input_frame_num: int = 1

    def reset(self, graph: BaseGraph = None) -> None:
        self._table = {}
        self._graph = graph

    def get_table(self) -> Dict[str, Dict[str, Any]]:
        return self._table

    def append(self, d) -> Dict[str, Dict[str, Any]]:
        if isinstance(d, dict):
            self._table.update(d)
        elif isinstance(d, list):
            for item in d:
                self._table.update(item)
        return self._table

    def add(self, var_name: str, op_name: str, window_size: int, frame_axis: int = None) -> None:
        self._table[var_name] = {
            'op_name': op_name,
            'window_size': window_size,
            'frame_axis': frame_axis if frame_axis is not None else self._input_frame_axis,
        }

    def check(self, graph: BaseGraph) -> bool:
        if graph is None:
            logger.error("Graph is None, check failed.")
            return False
        for var_name in self._table:
            if var_name not in graph.variables:
                logger.error(f"Variable {var_name} not found in graph.")
                return False
        logger.info("Streaming table check passed.")
        return True

    def print(self) -> None:
        print(json.dumps(self._table, indent=4))

    def set_input_frame_axis(self, axis: int) -> None:
        self._input_frame_axis = axis

    def set_input_frame_num(self, num: int) -> None:
        self._input_frame_num = num

    # ---------- add dict interface ----------
    def __contains__(self, var_name: str) -> bool:
        return var_name in self._table

    def __setitem__(self, key: str, value: Dict[str, Any]) -> None:
        self._table[key] = value

    def __getitem__(self, key: str) -> Dict[str, Any]:
        return self._table[key]

    def __delitem__(self, key: str) -> None:
        del self._table[key]

    def __len__(self) -> int:
        return len(self._table)

    def __iter__(self):
        return iter(self._table)


def insert_streaming_cache_op(graph: BaseGraph, var_name: str, attrs: Dict[str, Any]) -> Operation:
    """
    Insert a Streaming Node on given variable, according to given perm.
    """
    window_size = attrs['window_size']
    op_name = attrs.get('op_name', None)
    frame_axis = attrs.get('frame_axis', 1)
    if window_size <= 1:
        logger.info(f"Bypass inserting StreamingCache on Variable {var_name} due to window_size <= 1.")
        return None
    created = None
    op = None
    var = graph.variables[var_name]
    var_input_index = None
    var_output_index = None
    info = ExporterPatternInfo()

    if op_name:
        op = graph.operations[op_name]
        created_op_name = f"{var_name}_{op_name}_SCache"
        if var in op.inputs:
            var_input_index = op.inputs.index(var)
        elif var in op.outputs:
            var_output_index = op.outputs.index(var)
        logger.info(
            f"Inserting StreamingCache between Variable({var.name}) and Operation({op.name}): window_size {window_size}, frame_axis {frame_axis}"
        )
    elif len(var.dest_ops) > 0:
        op = var.dest_ops[0]
        created_op_name = f"{var_name}_SCache"
        var_input_index = op.inputs.index(var)
        logger.info(
            f"Inserting StreamingCache on Variable {var.name}: window_size {window_size}, frame_axis {frame_axis}"
        )
    elif var.source_op != None:
        op = var.source_op
        created_op_name = f"{var_name}_SCache"
        var_output_index = op.outputs.index(var)
        logger.info(
            f"Inserting StreamingCache on Variable {var.name}: window_size {window_size}, frame_axis {frame_axis}"
        )
    else:
        raise ValueError(f"Variable {var.name} has no connected operations.")

    created = graph.create_operation(
        op_type="StreamingCache",
        attributes={"window_size": window_size, "frame_axis": frame_axis},
        name=created_op_name,
    )
    if isinstance(op, QuantableOperation):
        # For StreamingCache op,  input_quantization_config == output_quantization_config
        if var_input_index != None:
            new_config = OperationQuantizationConfig(
                [op.input_quant_config[var_input_index]],
                [op.input_quant_config[var_input_index]],
            )
        else:
            new_config = OperationQuantizationConfig(
                [op.output_quant_config[var_output_index]],
                [op.output_quant_config[var_output_index]],
            )
        created = QuantableOperation(created, new_config, op.platform)
        created.attributes["quant_type"] = op.attributes["quant_type"]
        graph.operations[created.name] = created

    if op_name:
        if var_input_index != None:
            graph.insert_op_before(A=created, B=op, input_idx=var_input_index)
        elif var_output_index != None:
            graph.insert_op_after(A=created, B=op, output_idx=var_output_index)
    else:
        graph.insert_op_on_var(created, var_name)

    new_var = created.outputs[0]
    new_var.shape = copy.deepcopy(var.shape)
    perm = info.get_var_permute(var_name)
    frame_axis_perm = perm.index(frame_axis)
    new_var.shape[frame_axis_perm] = window_size + var.shape[frame_axis_perm] - 1
    new_var.is_parameter = False
    new_var.dtype = var.dtype
    info.add_var_exponents(new_var.name, info.get_var_exponents(var_name))
    info.add_var_permute(new_var.name, perm)

    return created


def set_conv_new_padding(op: Operation) -> None:
    padding = op.attributes.get('pads', None)

    if padding != None:
        if len(padding) == 2:
            new_padding = [0, 0]
        elif len(padding) == 4:
            new_padding = [0, padding[1], 0, padding[3]]
        op.attributes['pads'] = new_padding


def get_conv_cache_window_size(op: Operation) -> int:
    """
    Get suitable window size for convolution operation.
    """
    if op.type not in CONV_STREAMING_OP_SET:
        raise ValueError(f"Operation {op.name} is not a convolution operation.")

    # Assuming the first input is the feature map and the second input is the kernel
    input_var = op.inputs[0]
    if len(op.inputs) > 1:
        kernel_shape = op.inputs[1].shape
    elif 'kernel_shape' in op.attributes:
        kernel_shape = op.attributes['kernel_shape']
    else:
        raise ValueError(f"Cannot determine kernel shape for operation {op.name}.")

    window_size = 0
    kernel_height = kernel_shape[0]
    padding = op.attributes.get('pads', None)
    stride = op.attributes.get('strides', [1])
    dilation = op.attributes.get('dilations', [1])
    auto_pad = op.attributes.get('auto_pad', 'NOTSET')
    effective_kernel_size = dilation[0] * (kernel_height - 1) + 1

    if auto_pad in ['SAME_UPPER', 'SAME_LOWER'] and effective_kernel_size > 1:
        # Handle auto padding cases if necessary
        logger.error(f"Auto padding '{auto_pad}' is not supported in streaming node.")
        return window_size

    if padding != None:
        if len(padding) == 2:
            pad_bottom = padding[1]
        elif len(padding) == 4:
            pad_bottom = padding[2]
    else:
        pad_bottom = 0

    if pad_bottom > 0:
        logger.error(f"{padding} is not supported in streaming node.")
        return window_size

    window_size = effective_kernel_size

    return window_size


class AutoStreamingPattern(OperationExporter):
    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        streaming_table = StreamingTable()

        if op.type in CONV_STREAMING_OP_SET:
            input_var = op.inputs[0]
            if input_var.name in streaming_table:
                logger.info(f"Bypass streaming node for op {op.name} as it is already in streaming table.")
            else:
                window_size = get_conv_cache_window_size(op)
                if window_size > 1:
                    streaming_table.add(input_var.name, op.name, window_size)
                    logger.info(
                        f"Insert streaming node for op {op.name} with window size {window_size} on variable {input_var.name}."
                    )
                elif window_size == 1:
                    logger.info(f"Bypass streaming node for op {op.name} due to window size == 1.")
                else:
                    logger.error(f"Failed to insert streaming node for op {op.name} due to invalid window size.")
            set_conv_new_padding(op)
        elif op.type in STREADMING_BYPASS_OP_SET:
            logger.info(f"Bypass streaming node for op {op.name} of type {op.type}.")
        else:
            manual_flag = False
            for var in op.inputs:
                if var.name in streaming_table:
                    logger.info(
                        f"Bypass streaming node for op {op.name} as input var {var.name} is already in streaming table."
                    )
                    manual_flag = True
            if not manual_flag:
                logger.error(
                    f"Operation {op.name} of type {op.type} is not supported for AutoStreamingPattern. Please insert manually on variables: {[var.name for var in op.inputs]}."
                )

        return op


def insert_streaming_nodes(graph: BaseGraph):
    """
    Reset layout from NCHW -> NHWC
    """

    streaming_table = StreamingTable()
    streaming_table.check(graph)

    for var_name, attrs in streaming_table.get_table().items():
        insert_streaming_cache_op(graph, var_name, attrs)



# ---------------------------------------------------------------
# 这个脚本向你展示了如何使用 openvino 对 PPQ 导出的模型进行推理
# 你需要注意，openvino 也可以运行各种各样的量化方案，你甚至可以用 tensorRT 的 policy
# 但总的来说，openvino 需要非对称量化的 activation 和对称量化的 weights
# 现在的写法针对单输入网络哦，多输入的你得自己改改
# ---------------------------------------------------------------

# For this onnx inference test, all test data is randomly picked.
# If you want to use real data, just rewrite the defination of SAMPLES
import time

import numpy as np
import openvino
import torch
from tqdm import tqdm

from esp_ppq import *
from esp_ppq.api import *

QUANT_PLATFROM = TargetPlatform.OPENVINO_INT8
BATCHSIZE = 1
DEVICE = 'cuda'
INPUTSHAPE = [BATCHSIZE, 3, 640, 640]
SAMPLES = [torch.rand(size=INPUTSHAPE) for _ in range(256)]
BENCHMARK_SAMPLES = 512
MODEL_PATH = 'Models/yolox_s.onnx'
VALIDATION = False

with ENABLE_CUDA_KERNEL():
    quantized = quantize_onnx_model(
        onnx_import_file=MODEL_PATH,
        calib_dataloader=SAMPLES,
        collate_fn=lambda x: x.to(DEVICE),
        calib_steps=32,
        input_shape=INPUTSHAPE,
        setting=QuantizationSettingFactory.default_setting(),
        platform=QUANT_PLATFROM,
    )

    graphwise_error_analyse(
        graph=quantized, running_device='cuda', dataloader=SAMPLES, collate_fn=lambda x: x.cuda(), steps=32
    )

    export_ppq_graph(graph=quantized, platform=TargetPlatform.ONNX, graph_save_to='FP32.onnx')

    export_ppq_graph(graph=quantized, platform=TargetPlatform.OPENVINO_INT8, graph_save_to='INT8.onnx')

from esp_ppq.utils.OpenvinoUtil import Benchmark

Benchmark(ir_or_onnx_file='FP32.onnx', samples=500, jobs=4)
Benchmark(ir_or_onnx_file='INT8.onnx', samples=500, jobs=4)
